""" Defines the model architecture of SPGNN.
"""

import torch
import torch.nn.functional as F
from torch_geometric.nn import (
    GlobalAttention,
    MessagePassing,
    Set2Set,
    global_add_pool,
    global_max_pool,
    global_mean_pool,
)
from torch_geometric.nn.inits import glorot, zeros
from torch_geometric.utils import add_self_loops, degree, softmax
from torch_scatter import scatter_add

from olorenchemengine.internal import mock_imports

num_atom_type = 120  # including the extra mask tokens
num_chirality_tag = 3

num_bond_type = 6  # including aromatic and self-loop edge, and extra masked tokens
num_bond_direction = 3


class GINConv(MessagePassing):
    """
    Extension of GIN aggregation to incorporate edge information by concatenation.
    Args:
        emb_dim (int): dimensionality of embeddings for nodes and edges.
        embed_input (bool): whether to embed input or not.

    See https://arxiv.org/abs/1810.00826
    """

    def __init__(self, emb_dim, aggr="add"):
        super(GINConv, self).__init__(aggr=aggr)
        # multi-layer perceptron
        self.mlp = torch.nn.Sequential(
            torch.nn.Linear(emb_dim, 2 * emb_dim),
            torch.nn.ReLU(),
            torch.nn.Linear(2 * emb_dim, emb_dim),
        )
        self.edge_embedding1 = torch.nn.Embedding(num_bond_type, emb_dim)
        self.edge_embedding2 = torch.nn.Embedding(num_bond_direction, emb_dim)

        torch.nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        torch.nn.init.xavier_uniform_(self.edge_embedding2.weight.data)
        self.aggr = aggr

    def forward(self, x, edge_index, edge_attr):
        # add self loops in the edge space
        edge_index, _ = add_self_loops(edge_index, num_nodes=x.size(0))

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:, 0] = 4  # bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        edge_embeddings = self.edge_embedding1(edge_attr[:, 0]) + self.edge_embedding2(
            edge_attr[:, 1]
        )

        return self.propagate(edge_index=edge_index, x=x, edge_attr=edge_embeddings)

    def message(self, x_j, edge_attr):
        return x_j + edge_attr

    def update(self, aggr_out):
        return self.mlp(aggr_out)


class GCNConv(MessagePassing):
    def __init__(self, emb_dim, aggr="add"):
        super(GCNConv, self).__init__()

        self.emb_dim = emb_dim
        self.linear = torch.nn.Linear(emb_dim, emb_dim)
        self.edge_embedding1 = torch.nn.Embedding(num_bond_type, emb_dim)
        self.edge_embedding2 = torch.nn.Embedding(num_bond_direction, emb_dim)

        torch.nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        torch.nn.init.xavier_uniform_(self.edge_embedding2.weight.data)

        self.aggr = aggr

    def norm(self, edge_index, num_nodes, dtype):
        ### assuming that self-loops have been already added in edge_index
        edge_weight = torch.ones(
            (edge_index.size(1),), dtype=dtype, device=edge_index.device
        )
        row, col = edge_index
        deg = scatter_add(edge_weight, row, dim=0, dim_size=num_nodes)
        deg_inv_sqrt = deg.pow(-0.5)
        deg_inv_sqrt[deg_inv_sqrt == float("inf")] = 0

        return deg_inv_sqrt[row] * edge_weight * deg_inv_sqrt[col]

    def forward(self, x, edge_index, edge_attr):
        # add self loops in the edge space
        edge_index = add_self_loops(edge_index, num_nodes=x.size(0))

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:, 0] = 4  # bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        edge_embeddings = self.edge_embedding1(edge_attr[:, 0]) + self.edge_embedding2(
            edge_attr[:, 1]
        )

        norm = self.norm(edge_index, x.size(0), x.dtype)

        x = self.linear(x)

        return self.propagate(
            self.aggr, edge_index, x=x, edge_attr=edge_embeddings, norm=norm
        )

    def message(self, x_j, edge_attr, norm):
        return norm.view(-1, 1) * (x_j + edge_attr)


class GATConv(MessagePassing):
    def __init__(self, emb_dim, heads=2, negative_slope=0.2, aggr="add"):
        super(GATConv, self).__init__()

        self.aggr = aggr

        self.emb_dim = emb_dim
        self.heads = heads
        self.negative_slope = negative_slope

        self.weight_linear = torch.nn.Linear(emb_dim, heads * emb_dim)
        self.att = torch.nn.Parameter(torch.Tensor(1, heads, 2 * emb_dim))

        self.bias = torch.nn.Parameter(torch.Tensor(emb_dim))

        self.edge_embedding1 = torch.nn.Embedding(num_bond_type, heads * emb_dim)
        self.edge_embedding2 = torch.nn.Embedding(num_bond_direction, heads * emb_dim)

        torch.nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        torch.nn.init.xavier_uniform_(self.edge_embedding2.weight.data)

        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.att)
        zeros(self.bias)

    def forward(self, x, edge_index, edge_attr):
        # add self loops in the edge space
        edge_index, _ = add_self_loops(edge_index, num_nodes=x.size(0))

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:, 0] = 4  # bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        edge_embeddings = self.edge_embedding1(edge_attr[:, 0]) + self.edge_embedding2(
            edge_attr[:, 1]
        )

        # x = self.weight_linear(x).view(-1, self.heads, self.emb_dim)
        x = self.weight_linear(x)
        # return self.propagate(self.aggr, edge_index[0], x=x, edge_attr=edge_embeddings)
        return self.propagate(edge_index, x=x, edge_attr=edge_embeddings)

    def message(self, edge_index, x_i, x_j, edge_attr):
        x_i = x_i.view(-1, self.heads, self.emb_dim)
        x_j = x_j.view(-1, self.heads, self.emb_dim)
        edge_attr = edge_attr.view(-1, self.heads, self.emb_dim)
        x_j += edge_attr

        alpha = (torch.cat([x_i, x_j], dim=-1) * self.att).sum(dim=-1)

        alpha = F.leaky_relu(alpha, self.negative_slope)
        alpha = softmax(alpha, edge_index[0])

        out = x_j * alpha.view(-1, self.heads, 1)
        out = out.view(-1, self.heads * self.emb_dim)
        return out

    def update(self, aggr_out):
        aggr_out = aggr_out.view(-1, self.heads, self.emb_dim).mean(dim=1)
        aggr_out = aggr_out + self.bias
        return aggr_out


class GraphSAGEConv(MessagePassing):
    def __init__(self, emb_dim, aggr="mean"):
        super(GraphSAGEConv, self).__init__()

        self.emb_dim = emb_dim
        self.linear = torch.nn.Linear(emb_dim, emb_dim)
        self.edge_embedding1 = torch.nn.Embedding(num_bond_type, emb_dim)
        self.edge_embedding2 = torch.nn.Embedding(num_bond_direction, emb_dim)

        torch.nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        torch.nn.init.xavier_uniform_(self.edge_embedding2.weight.data)

        self.aggr = aggr

    def forward(self, x, edge_index, edge_attr):
        # add self loops in the edge space
        edge_index = add_self_loops(edge_index, num_nodes=x.size(0))

        # add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:, 0] = 4  # bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim=0)

        edge_embeddings = self.edge_embedding1(edge_attr[:, 0]) + self.edge_embedding2(
            edge_attr[:, 1]
        )

        x = self.linear(x)

        return self.propagate(self.aggr, edge_index, x=x, edge_attr=edge_embeddings)

    def message(self, x_j, edge_attr):
        return x_j + edge_attr

    def update(self, aggr_out):
        return F.normalize(aggr_out, p=2, dim=-1)


class GNN(torch.nn.Module):
    """

    Args:
        num_layer (int): the number of GNN layers
        emb_dim (int): dimensionality of embeddings
        JK (str): last, concat, max or sum.
        max_pool_layer (int): the layer from which we use max pool rather than add pool for neighbor aggregation
        drop_ratio (float): dropout rate
        gnn_type: gin, gcn, graphsage, gat
    Output:
        node representations
    """

    def __init__(self, num_layer, emb_dim, JK="last", drop_ratio=0, gnn_type="gin"):
        super(GNN, self).__init__()
        self.num_layer = num_layer
        self.drop_ratio = drop_ratio
        self.JK = JK

        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.x_embedding1 = torch.nn.Embedding(num_atom_type, emb_dim)
        self.x_embedding2 = torch.nn.Embedding(num_chirality_tag, emb_dim)

        torch.nn.init.xavier_uniform_(self.x_embedding1.weight.data)
        torch.nn.init.xavier_uniform_(self.x_embedding2.weight.data)

        ###List of MLPs
        self.gnns = torch.nn.ModuleList()
        for layer in range(num_layer):
            if gnn_type == "gin":
                self.gnns.append(GINConv(emb_dim, aggr="add"))
            elif gnn_type == "gcn":
                self.gnns.append(GCNConv(emb_dim))
            elif gnn_type == "gat":
                self.gnns.append(GATConv(emb_dim))
            elif gnn_type == "graphsage":
                self.gnns.append(GraphSAGEConv(emb_dim))

        ###List of batchnorms
        self.batch_norms = torch.nn.ModuleList()
        for layer in range(num_layer):
            self.batch_norms.append(torch.nn.BatchNorm1d(emb_dim))

    # def forward(self, x, edge_index, edge_attr):
    def forward(self, *argv):
        if len(argv) == 3:
            x, edge_index, edge_attr = argv[0], argv[1], argv[2]
        elif len(argv) == 1:
            data = argv[0]
            x, edge_index, edge_attr = data.x, data.edge_index, data.edge_attr
        else:
            raise ValueError("unmatched number of arguments.")

        x = self.x_embedding1(x[:, 0]) + self.x_embedding2(x[:, 1])

        h_list = [x]
        for layer in range(self.num_layer):
            h = self.gnns[layer](h_list[layer], edge_index, edge_attr)
            h = self.batch_norms[layer](h)
            # h = F.dropout(F.relu(h), self.drop_ratio, training = self.training)
            if layer == self.num_layer - 1:
                # remove relu for the last layer
                h = F.dropout(h, self.drop_ratio, training=self.training)
            else:
                h = F.dropout(F.relu(h), self.drop_ratio, training=self.training)
            h_list.append(h)

        ### Different implementations of Jk-concat
        if self.JK == "concat":
            node_representation = torch.cat(h_list, dim=1)
        elif self.JK == "last":
            node_representation = h_list[-1]
        elif self.JK == "max":
            h_list = [h.unsqueeze_(0) for h in h_list]
            node_representation = torch.max(torch.cat(h_list, dim=0), dim=0)[0]
        elif self.JK == "sum":
            h_list = [h.unsqueeze_(0) for h in h_list]
            node_representation = torch.sum(torch.cat(h_list, dim=0), dim=0)[0]

        return node_representation


class GNN_graphpred(torch.nn.Module):
    """
    Extension of GIN to incorporate edge information by concatenation.
    Args:
        num_layer (int): the number of GNN layers
        emb_dim (int): dimensionality of embeddings
        num_tasks (int): number of tasks in multi-task learning scenario
        drop_ratio (float): dropout rate
        JK (str): last, concat, max or sum.
        graph_pooling (str): sum, mean, max, attention, set2set
        gnn_type: gin, gcn, graphsage, gat

    See https://arxiv.org/abs/1810.00826
    JK-net: https://arxiv.org/abs/1806.03536
    """

    def __init__(
        self,
        num_layer,
        emb_dim,
        num_tasks,
        JK="last",
        drop_ratio=0,
        graph_pooling="mean",
        gnn_type="gin",
    ):
        super(GNN_graphpred, self).__init__()
        self.num_layer = num_layer
        self.drop_ratio = drop_ratio
        self.JK = JK
        self.emb_dim = emb_dim
        self.num_tasks = num_tasks

        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.gnn = GNN(num_layer, emb_dim, JK, drop_ratio, gnn_type=gnn_type)

        # Different kind of graph pooling
        if graph_pooling == "sum":
            self.pool = global_add_pool
        elif graph_pooling == "mean":
            self.pool = global_mean_pool
        elif graph_pooling == "max":
            self.pool = global_max_pool
        elif graph_pooling == "attention":
            if self.JK == "concat":
                self.pool = GlobalAttention(
                    gate_nn=torch.nn.Linear((self.num_layer + 1) * emb_dim, 1)
                )
            else:
                self.pool = GlobalAttention(gate_nn=torch.nn.Linear(emb_dim, 1))
        elif graph_pooling[:-1] == "set2set":
            set2set_iter = int(graph_pooling[-1])
            if self.JK == "concat":
                self.pool = Set2Set((self.num_layer + 1) * emb_dim, set2set_iter)
            else:
                self.pool = Set2Set(emb_dim, set2set_iter)
        else:
            raise ValueError("Invalid graph pooling type.")

        # For graph-level binary classification
        if graph_pooling[:-1] == "set2set":
            self.mult = 2
        else:
            self.mult = 1

        if self.JK == "concat":
            self.graph_pred_linear = torch.nn.Linear(
                self.mult * (self.num_layer + 1) * self.emb_dim, self.num_tasks
            )
        else:
            self.graph_pred_linear = torch.nn.Linear(
                self.mult * self.emb_dim, self.num_tasks
            )

    def from_pretrained(self, model_file, map_location="cpu"):
        self.gnn.load_state_dict(torch.load(model_file, map_location=map_location))

    def forward(self, *argv):
        if len(argv) == 4:
            x, edge_index, edge_attr, batch = argv[0], argv[1], argv[2], argv[3]
        elif len(argv) == 1:
            data = argv[0]
            x, edge_index, edge_attr, batch = (
                data.x,
                data.edge_index,
                data.edge_attr,
                data.batch,
            )
        else:
            raise ValueError("unmatched number of arguments.")

        node_representation = self.gnn(x, edge_index, edge_attr)

        return self.graph_pred_linear(self.pool(node_representation, batch))



from typing import *

from olorenchemengine.base_class import *
from olorenchemengine.dataset import *
from olorenchemengine.visualizations.visualization import *


class MatchedPairsTable(BaseVisualization):
    @log_arguments
    def __init__(
        self,
        dataset: BaseDataset,
        mode: str = "features",
        annotations: Union[str, List[str]] = [],
    ):

        self.packages = ["d3"]

        self.dataset = dataset

        if not isinstance(annotations, list):
            annotations = [annotations]
        self.annotations = annotations

        self.mps = self.matched_pairs(dataset.data, dataset.feature_cols)
        if mode == "features":
            self.mp_diffs = pd.DataFrame()

            for mp in self.mps:
                r1 = self.dataset.data.iloc[mp[0]]
                r2 = self.dataset.data.iloc[mp[1]]
                diff = mp[2][0]
                self.mp_diffs = self.mp_diffs.append(
                    {
                        "col": diff[0],
                        "initial": diff[1],
                        "final": diff[2],
                        **{
                            f"Delta (Final - Initial) {property_col}": r2[
                                property_col + " mean"
                            ]
                            - r1[property_col + " mean"]
                            for property_col in dataset.property_col
                        },
                        **{
                            f"Fold (Final/Initial) {property_col}": r2[
                                property_col + " mean"
                            ]
                            / r1[property_col + " mean"]
                            for property_col in dataset.property_col
                        },
                        **{
                            f"Initial {annotation}": r1[annotation]
                            for annotation in self.annotations
                        },
                        **{
                            f"Final {annotation}": r2[annotation]
                            for annotation in self.annotations
                        },
                    },
                    ignore_index=True,
                )
            self.mp_diffs = (
                self.mp_diffs.sort_values(by="col", ascending=True)
                .reset_index()
                .drop(columns=["index"])
            )
        elif mode == "property":
            self.mp_diffs = pd.DataFrame()
            for mp in self.mps:
                r1 = self.dataset.data.iloc[mp[0]]
                r2 = self.dataset.data.iloc[mp[1]]
                diff = mp[2][0]

                for property in self.dataset.property_col:
                    if not r1[property] == r2[property]:
                        self.mp_diffs = self.mp_diffs.append(
                            {
                                "col": diff[0],
                                "initial": diff[1],
                                "final": diff[2],
                                f"Initial {property}": r1[property],
                                f"Final {property}": r2[property],
                                **{
                                    f"Initial {annotation}": r1[annotation]
                                    for annotation in annotations
                                },
                                **{
                                    f"Final {annotation}": r2[annotation]
                                    for annotation in annotations
                                },
                            },
                            ignore_index=True,
                        )
            self.mp_diffs = (
                self.mp_diffs.sort_values(
                    by=["col", "initial", "final"], ascending=True
                )
                .reset_index()
                .drop(columns=["index"])
            )

    def matched_pairs(self, df, cols, dist=1):
        out = []
        for i_ in tqdm(range(len(df.index))):
            i = df.index[i_]
            for j_ in range(len(df.index)):
                j = df.index[j_]
                if not i_ == j_:
                    count = []
                    close = True
                    for col in cols:
                        if str(df.iloc[i][col]) != str(df.iloc[j][col]):
                            count.append((col, df.iloc[i][col], df.iloc[j][col]))
                        if len(count) > dist:
                            close = False
                            break
                    if close:
                        out.append((i, j, count))
        return out

    def get_data(self):
        data = []
        for i, col in enumerate(self.mp_diffs["col"].unique()):
            data.append(
                {
                    "col": col,
                    "i": i,
                    "col_data": self.mp_diffs.loc[self.mp_diffs["col"] == col].to_dict(
                        "r"
                    ),
                }
            )
        return data

    @property
    def JS_NAME(self):
        return "MatchedPairsTable"


class MatchedPairsHeatmap(MatchedPairsTable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.packages = ["d3", "plotly"]

    def get_data(self):
        data = []
        for property_col in self.dataset.property_col:
            property_col = f"Delta (Final - Initial) {property_col}"
            for i, col in enumerate(self.mp_diffs["col"].unique()):
                col_data = self.mp_diffs[self.mp_diffs["col"] == col]
                mapping = {
                    str(x): i for i, x in enumerate(col_data["initial"].unique())
                }
                col_heatmap = np.zeros((len(mapping), len(mapping)))
                for i, col_row in col_data.iterrows():
                    try:
                        x = float(col_row[property_col])
                        if not np.isnan(x):
                            col_heatmap[
                                mapping[str(col_row["final"])],
                                mapping[str(col_row["initial"])],
                            ] += col_row[property_col]
                    except:
                        pass

                data.append(
                    {
                        "z": col_heatmap.tolist(),
                        "x": list(mapping.keys()),
                        "y": list(mapping.keys()),
                        "xlabel": "Initial",
                        "ylabel": "Final",
                        "title": f"Residue number {col}, colored by sum of {property_col}",
                    }
                )
        return data

    @property
    def JS_NAME(self):
        return "MatchedPairsHeatmap"



#  COPYRIGHT Aspuru-Guzik group (2021)
#                                  Apache License
#                            Version 2.0, January 2004
#                         http://www.apache.org/licenses/

#    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

#    1. Definitions.

#       "License" shall mean the terms and conditions for use, reproduction,
#       and distribution as defined by Sections 1 through 9 of this document.

#       "Licensor" shall mean the copyright owner or entity authorized by
#       the copyright owner that is granting the License.

#       "Legal Entity" shall mean the union of the acting entity and all
#       other entities that control, are controlled by, or are under common
#       control with that entity. For the purposes of this definition,
#       "control" means (i) the power, direct or indirect, to cause the
#       direction or management of such entity, whether by contract or
#       otherwise, or (ii) ownership of fifty percent (50%) or more of the
#       outstanding shares, or (iii) beneficial ownership of such entity.

#       "You" (or "Your") shall mean an individual or Legal Entity
#       exercising permissions granted by this License.

#       "Source" form shall mean the preferred form for making modifications,
#       including but not limited to software source code, documentation
#       source, and configuration files.

#       "Object" form shall mean any form resulting from mechanical
#       transformation or translation of a Source form, including but
#       not limited to compiled object code, generated documentation,
#       and conversions to other media types.

#       "Work" shall mean the work of authorship, whether in Source or
#       Object form, made available under the License, as indicated by a
#       copyright notice that is included in or attached to the work
#       (an example is provided in the Appendix below).

#       "Derivative Works" shall mean any work, whether in Source or Object
#       form, that is based on (or derived from) the Work and for which the
#       editorial revisions, annotations, elaborations, or other modifications
#       represent, as a whole, an original work of authorship. For the purposes
#       of this License, Derivative Works shall not include works that remain
#       separable from, or merely link (or bind by name) to the interfaces of,
#       the Work and Derivative Works thereof.

#       "Contribution" shall mean any work of authorship, including
#       the original version of the Work and any modifications or additions
#       to that Work or Derivative Works thereof, that is intentionally
#       submitted to Licensor for inclusion in the Work by the copyright owner
#       or by an individual or Legal Entity authorized to submit on behalf of
#       the copyright owner. For the purposes of this definition, "submitted"
#       means any form of electronic, verbal, or written communication sent
#       to the Licensor or its representatives, including but not limited to
#       communication on electronic mailing lists, source code control systems,
#       and issue tracking systems that are managed by, or on behalf of, the
#       Licensor for the purpose of discussing and improving the Work, but
#       excluding communication that is conspicuously marked or otherwise
#       designated in writing by the copyright owner as "Not a Contribution."

#       "Contributor" shall mean Licensor and any individual or Legal Entity
#       on behalf of whom a Contribution has been received by Licensor and
#       subsequently incorporated within the Work.

#    2. Grant of Copyright License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       copyright license to reproduce, prepare Derivative Works of,
#       publicly display, publicly perform, sublicense, and distribute the
#       Work and such Derivative Works in Source or Object form.

#    3. Grant of Patent License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       (except as stated in this section) patent license to make, have made,
#       use, offer to sell, sell, import, and otherwise transfer the Work,
#       where such license applies only to those patent claims licensable
#       by such Contributor that are necessarily infringed by their
#       Contribution(s) alone or by combination of their Contribution(s)
#       with the Work to which such Contribution(s) was submitted. If You
#       institute patent litigation against any entity (including a
#       cross-claim or counterclaim in a lawsuit) alleging that the Work
#       or a Contribution incorporated within the Work constitutes direct
#       or contributory patent infringement, then any patent licenses
#       granted to You under this License for that Work shall terminate
#       as of the date such litigation is filed.

#    4. Redistribution. You may reproduce and distribute copies of the
#       Work or Derivative Works thereof in any medium, with or without
#       modifications, and in Source or Object form, provided that You
#       meet the following conditions:

#       (a) You must give any other recipients of the Work or
#           Derivative Works a copy of this License; and

#       (b) You must cause any modified files to carry prominent notices
#           stating that You changed the files; and

#       (c) You must retain, in the Source form of any Derivative Works
#           that You distribute, all copyright, patent, trademark, and
#           attribution notices from the Source form of the Work,
#           excluding those notices that do not pertain to any part of
#           the Derivative Works; and

#       (d) If the Work includes a "NOTICE" text file as part of its
#           distribution, then any Derivative Works that You distribute must
#           include a readable copy of the attribution notices contained
#           within such NOTICE file, excluding those notices that do not
#           pertain to any part of the Derivative Works, in at least one
#           of the following places: within a NOTICE text file distributed
#           as part of the Derivative Works; within the Source form or
#           documentation, if provided along with the Derivative Works; or,
#           within a display generated by the Derivative Works, if and
#           wherever such third-party notices normally appear. The contents
#           of the NOTICE file are for informational purposes only and
#           do not modify the License. You may add Your own attribution
#           notices within Derivative Works that You distribute, alongside
#           or as an addendum to the NOTICE text from the Work, provided
#           that such additional attribution notices cannot be construed
#           as modifying the License.

#       You may add Your own copyright statement to Your modifications and
#       may provide additional or different license terms and conditions
#       for use, reproduction, or distribution of Your modifications, or
#       for any such Derivative Works as a whole, provided Your use,
#       reproduction, and distribution of the Work otherwise complies with
#       the conditions stated in this License.

#    5. Submission of Contributions. Unless You explicitly state otherwise,
#       any Contribution intentionally submitted for inclusion in the Work
#       by You to the Licensor shall be under the terms and conditions of
#       this License, without any additional terms or conditions.
#       Notwithstanding the above, nothing herein shall supersede or modify
#       the terms of any separate license agreement you may have executed
#       with Licensor regarding such Contributions.

#    6. Trademarks. This License does not grant permission to use the trade
#       names, trademarks, service marks, or product names of the Licensor,
#       except as required for reasonable and customary use in describing the
#       origin of the Work and reproducing the content of the NOTICE file.

#    7. Disclaimer of Warranty. Unless required by applicable law or
#       agreed to in writing, Licensor provides the Work (and each
#       Contributor provides its Contributions) on an "AS IS" BASIS,
#       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#       implied, including, without limitation, any warranties or conditions
#       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
#       PARTICULAR PURPOSE. You are solely responsible for determining the
#       appropriateness of using or redistributing the Work and assume any
#       risks associated with Your exercise of permissions under this License.

#    8. Limitation of Liability. In no event and under no legal theory,
#       whether in tort (including negligence), contract, or otherwise,
#       unless required by applicable law (such as deliberate and grossly
#       negligent acts) or agreed to in writing, shall any Contributor be
#       liable to You for damages, including any direct, indirect, special,
#       incidental, or consequential damages of any character arising as a
#       result of this License or out of the use or inability to use the
#       Work (including but not limited to damages for loss of goodwill,
#       work stoppage, computer failure or malfunction, or any and all
#       other commercial damages or losses), even if such Contributor
#       has been advised of the possibility of such damages.

#    9. Accepting Warranty or Additional Liability. While redistributing
#       the Work or Derivative Works thereof, You may choose to offer,
#       and charge a fee for, acceptance of support, warranty, indemnity,
#       or other liability obligations and/or rights consistent with this
#       License. However, in accepting such obligations, You may act only
#       on Your own behalf and on Your sole responsibility, not on behalf
#       of any other Contributor, and only if You agree to indemnify,
#       defend, and hold each Contributor harmless for any liability
#       incurred by, or claims asserted against, such Contributor by reason
#       of your accepting any such warranty or additional liability.

#    END OF TERMS AND CONDITIONS

#    APPENDIX: How to apply the Apache License to your work.

#       To apply the Apache License to your work, attach the following
#       boilerplate notice, with the fields enclosed by brackets "[]"
#       replaced with your own identifying information. (Don't include
#       the brackets!)  The text should be enclosed in the appropriate
#       comment syntax for the file format. We also recommend that a
#       file or class name and description of purpose be included on the
#       same "printed page" as the copyright notice for easier
#       identification within third-party archives.

#    Copyright [yyyy] [name of copyright owner]

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

#   CHANGES
#   1. Simplified imports
#   2. Added optional passed in index
#   3. Added condition to sanitize SMILES

import numpy as np
from rdkit import Chem, RDLogger
from rdkit.Chem import AllChem, Mol
from rdkit.Chem.AtomPairs.Sheridan import GetBPFingerprint, GetBTFingerprint
from rdkit.Chem.Pharm2D import Generate, Gobbi_Pharm2D
from rdkit.DataStructs.cDataStructs import TanimotoSimilarity

import olorenchemengine as oce

RDLogger.DisableLog("rdApp.*")


def randomize_smiles(mol):
    """Returns a random (dearomatized) SMILES given an rdkit mol object of a molecule.
    Parameters:
    mol (rdkit.Chem.rdchem.Mol) :  RdKit mol object (None if invalid smile string smi)

    Returns:
    mol (rdkit.Chem.rdchem.Mol) : RdKit mol object  (None if invalid smile string smi)
    """
    if not mol:
        return None

    Chem.Kekulize(mol)
    return Chem.MolToSmiles(
        mol, canonical=False, doRandom=True, isomericSmiles=False, kekuleSmiles=True
    )


def sanitize_smiles(smi):
    """Return a canonical smile representation of smi

    Parameters:
    smi (string) : smile string to be canonicalized

    Returns:
    mol (rdkit.Chem.rdchem.Mol) : RdKit mol object                          (None if invalid smile string smi)
    smi_canon (string)          : Canonicalized smile representation of smi (None if invalid smile string smi)
    conversion_successful (bool): True/False to indicate if conversion was  successful
    """
    try:
        mol = Chem.MolFromSmiles(smi, sanitize=True)
        smi_canon = Chem.MolToSmiles(mol, isomericSmiles=False, canonical=True)
        if Chem.MolFromSmiles(smi_canon) is None:
            return (None, None, False)
        return (mol, smi_canon, True)
    except:
        return (None, None, False)


def get_selfie_chars(selfie):
    """Obtain a list of all selfie characters in string selfie

    Parameters:
    selfie (string) : A selfie string - representing a molecule

    Example:
    >>> get_selfie_chars('[C][=C][C][=C][C][=C][Ring1][Branch1_1]')
    ['[C]', '[=C]', '[C]', '[=C]', '[C]', '[=C]', '[Ring1]', '[Branch1_1]']

    Returns:
    chars_selfie: list of selfie characters present in molecule selfie
    """
    chars_selfie = []  # A list of all SELFIE sybols from string selfie
    while selfie != "":
        chars_selfie.append(selfie[selfie.find("[") : selfie.find("]") + 1])
        selfie = selfie[selfie.find("]") + 1 :]
    return chars_selfie


def mutate_selfie(selfie, max_molecules_len, write_fail_cases=False, index=None):
    """Return a mutated selfie string (only one mutation on selfie is performed)

    Mutations are done until a valid molecule is obtained
    Rules of mutation: With a 33.3% propbabily, either:
        1. Add a random SELFIE character in the string
        2. Replace a random SELFIE character with another
        3. Delete a random character

    Parameters:
    selfie            (string)  : SELFIE string to be mutated
    max_molecules_len (int)     : Mutations of SELFIE string are allowed up to this length
    write_fail_cases  (bool)    : If true, failed mutations are recorded in "selfie_failure_cases.txt"

    Returns:
    selfie_mutated    (string)  : Mutated SELFIE string
    smiles_canon      (string)  : canonical smile of mutated SELFIE string
    """
    valid = False
    fail_counter = 0
    chars_selfie = get_selfie_chars(selfie)

    while not valid:
        fail_counter += 1

        oce.import_or_install("selfies")

        import selfies

        alphabet = list(selfies.get_semantic_robust_alphabet())  # 34 SELFIE characters

        choice_ls = [1, 2, 3]  # 1=Insert; 2=Replace; 3=Delete
        random_choice = np.random.choice(choice_ls, 1)[0]

        # Insert a character in a Random Location
        if random_choice == 1:
            if index is None:
                random_index = np.random.randint(len(chars_selfie) + 1)
            else:
                random_index = index
            random_character = np.random.choice(alphabet, size=1)[0]

            selfie_mutated_chars = (
                chars_selfie[:random_index]
                + [random_character]
                + chars_selfie[random_index:]
            )

        # Replace a random character
        elif random_choice == 2:
            if index is None:
                random_index = np.random.randint(len(chars_selfie))
            else:
                random_index = index
            random_character = np.random.choice(alphabet, size=1)[0]
            if random_index == 0:
                selfie_mutated_chars = [random_character] + chars_selfie[
                    random_index + 1 :
                ]
            else:
                selfie_mutated_chars = (
                    chars_selfie[:random_index]
                    + [random_character]
                    + chars_selfie[random_index + 1 :]
                )

        # Delete a random character
        elif random_choice == 3:
            if index is None:
                random_index = np.random.randint(len(chars_selfie))
            else:
                random_index = index
            if random_index == 0:
                selfie_mutated_chars = chars_selfie[random_index + 1 :]
            else:
                selfie_mutated_chars = (
                    chars_selfie[:random_index] + chars_selfie[random_index + 1 :]
                )

        else:
            raise Exception("Invalid Operation trying to be performed")

        selfie_mutated = "".join(x for x in selfie_mutated_chars)
        sf = "".join(x for x in chars_selfie)

        try:
            smiles = decoder(selfie_mutated)
            mol, smiles_canon, done = sanitize_smiles(smiles)
            if len(selfie_mutated_chars) > max_molecules_len or smiles_canon == "":
                done = False
            if done:
                valid = True
            else:
                valid = False
        except:
            valid = False
            if fail_counter > 1 and write_fail_cases == True:
                f = open("selfie_failure_cases.txt", "a+")
                f.write(
                    "Tried to mutate SELFIE: "
                    + str(sf)
                    + " To Obtain: "
                    + str(selfie_mutated)
                    + "\n"
                )
                f.close()

    return (selfie_mutated, smiles_canon)


def get_mutated_SELFIES(selfies_ls, num_mutations):
    """Mutate all the SELFIES in 'selfies_ls' 'num_mutations' number of times.

    Parameters:
    selfies_ls   (list)  : A list of SELFIES
    num_mutations (int)  : number of mutations to perform on each SELFIES within 'selfies_ls'

    Returns:
    selfies_ls   (list)  : A list of mutated SELFIES

    """
    for _ in range(num_mutations):
        selfie_ls_mut_ls = []
        for str_ in selfies_ls:

            str_chars = get_selfie_chars(str_)
            max_molecules_len = len(str_chars) + num_mutations

            selfie_mutated, _ = mutate_selfie(str_, max_molecules_len)
            selfie_ls_mut_ls.append(selfie_mutated)

        selfies_ls = selfie_ls_mut_ls.copy()
    return selfies_ls



"""
Wrapping the GIN network family based on:

`GINEPLUS GitHub repository <https://github.com/RBrossard/GINEPLUS>`_
`Graph convolutions that can finally model local structure <https://arxiv.org/abs/2011.15069>`_
"""
from olorenchemengine.base_class import log_arguments
from olorenchemengine.gnn import BaseLightningModule

from .operations import GINNetwork


class GINModel(BaseLightningModule):

    """
    GINModel class from `GINEPLUS GitHub repository <https://github.com/RBrossard/GINEPLUS>`.

    Parameters:
        task_type (str): Type of task to perform. Options are 'regression' or 'classification'. This
            task_type will automatically be set by BaseTorchGeometricModel and usually
            does not need to be set manually.
        hidden (int): Number of nodes per hidden layer.
        lr (float): Learning rate.
        layers (int): Number of hidden layers.
        dropout (float): Dropout rate.
        virtual_node (bool): Whether to use virtual node.
        conv_type (str): Type of convolution to use. Options are 'gcn', 'gin', 'gin+', 'naivegin+'.
        conv_radius (int): Radius of convolution.
        optim (str): Optimizer to use. Options are 'adam'.
    """

    @log_arguments
    def __init__(
        self,
        task_type="classification",
        hidden: str = 100,
        lr: float = 0.001,
        layers=3,
        dropout=0.5,
        virtual_node=False,
        conv_radius=3,
        conv_type="gin+",
        optim="adam",
        **kwargs
    ):
        assert conv_type in ["gcn", "gin", "gin+", "naivegin+"]
        super().__init__(optim=optim)

        # Network
        out_dim = 1
        self.network = GINNetwork(
            hidden=hidden,
            out_dim=out_dim,
            layers=layers,
            dropout=dropout,
            virtual_node=virtual_node,
            k=conv_radius,
            conv_type=conv_type,
        )



import torch
import torch.nn.functional as F
from ogb.graphproppred.mol_encoder import AtomEncoder, BondEncoder

### GIN convolution along the graph structure
from torch_geometric.nn import (
    GlobalAttention,
    MessagePassing,
    Set2Set,
    global_add_pool,
    global_max_pool,
    global_mean_pool,
)
from torch_geometric.utils import degree


class GINConv(MessagePassing):
    def __init__(self, emb_dim):
        """
        emb_dim (int): node embedding dimensionality
        """

        super(GINConv, self).__init__(aggr="add")

        self.mlp = torch.nn.Sequential(
            torch.nn.Linear(emb_dim, 2 * emb_dim),
            torch.nn.BatchNorm1d(2 * emb_dim),
            torch.nn.ReLU(),
            torch.nn.Linear(2 * emb_dim, emb_dim),
        )
        self.eps = torch.nn.Parameter(torch.Tensor([0]))

        self.bond_encoder = BondEncoder(emb_dim=emb_dim)

    def forward(self, x, edge_index, edge_attr):
        edge_embedding = self.bond_encoder(edge_attr)
        out = self.mlp(
            (1 + self.eps) * x
            + self.propagate(edge_index, x=x, edge_attr=edge_embedding)
        )

        return out

    def message(self, x_j, edge_attr):
        return F.relu(x_j + edge_attr)

    def update(self, aggr_out):
        return aggr_out


### GCN convolution along the graph structure
class GCNConv(MessagePassing):
    def __init__(self, emb_dim):
        super(GCNConv, self).__init__(aggr="add")

        self.linear = torch.nn.Linear(emb_dim, emb_dim)
        self.root_emb = torch.nn.Embedding(1, emb_dim)
        self.bond_encoder = BondEncoder(emb_dim=emb_dim)

    def forward(self, x, edge_index, edge_attr):
        x = self.linear(x)
        edge_embedding = self.bond_encoder(edge_attr)

        row, col = edge_index

        # edge_weight = torch.ones((edge_index.size(1), ), device=edge_index.device)
        deg = degree(row, x.size(0), dtype=x.dtype) + 1
        deg_inv_sqrt = deg.pow(-0.5)
        deg_inv_sqrt[deg_inv_sqrt == float("inf")] = 0

        norm = deg_inv_sqrt[row] * deg_inv_sqrt[col]

        return self.propagate(
            edge_index, x=x, edge_attr=edge_embedding, norm=norm
        ) + F.relu(x + self.root_emb.weight) * 1.0 / deg.view(-1, 1)

    def message(self, x_j, edge_attr, norm):
        return norm.view(-1, 1) * F.relu(x_j + edge_attr)

    def update(self, aggr_out):
        return aggr_out


### GNN to generate node embedding
class GNN_node(torch.nn.Module):
    """
    Output:
        node representations
    """

    def __init__(
        self,
        num_layer,
        emb_dim,
        drop_ratio=0.5,
        JK="last",
        residual=False,
        gnn_type="gin",
    ):
        """
        emb_dim (int): node embedding dimensionality
        num_layer (int): number of GNN message passing layers
        """

        super(GNN_node, self).__init__()
        self.num_layer = num_layer
        self.drop_ratio = drop_ratio
        self.JK = JK
        ### add residual connection or not
        self.residual = residual

        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.atom_encoder = AtomEncoder(emb_dim)

        ###List of GNNs
        self.convs = torch.nn.ModuleList()
        self.batch_norms = torch.nn.ModuleList()

        for layer in range(num_layer):
            if gnn_type == "gin":
                self.convs.append(GINConv(emb_dim))
            elif gnn_type == "gcn":
                self.convs.append(GCNConv(emb_dim))
            else:
                raise ValueError("Undefined GNN type called {}".format(gnn_type))

            self.batch_norms.append(torch.nn.BatchNorm1d(emb_dim))

    def forward(self, batched_data):
        x, edge_index, edge_attr, batch = (
            batched_data.x,
            batched_data.edge_index,
            batched_data.edge_attr,
            batched_data.batch,
        )

        ### computing input node embedding

        h_list = [self.atom_encoder(x)]
        for layer in range(self.num_layer):

            h = self.convs[layer](h_list[layer], edge_index, edge_attr)
            h = self.batch_norms[layer](h)

            if layer == self.num_layer - 1:
                # remove relu for the last layer
                h = F.dropout(h, self.drop_ratio, training=self.training)
            else:
                h = F.dropout(F.relu(h), self.drop_ratio, training=self.training)

            if self.residual:
                h += h_list[layer]

            h_list.append(h)

        ### Different implementations of Jk-concat
        if self.JK == "last":
            node_representation = h_list[-1]
        elif self.JK == "sum":
            node_representation = 0
            for layer in range(self.num_layer + 1):
                node_representation += h_list[layer]

        return node_representation


### Virtual GNN to generate node embedding
class GNN_node_Virtualnode(torch.nn.Module):
    """
    Output:
        node representations
    """

    def __init__(
        self,
        num_layer,
        emb_dim,
        drop_ratio=0.5,
        JK="last",
        residual=False,
        gnn_type="gin",
    ):
        """
        emb_dim (int): node embedding dimensionality
        """

        super(GNN_node_Virtualnode, self).__init__()
        self.num_layer = num_layer
        self.drop_ratio = drop_ratio
        self.JK = JK
        ### add residual connection or not
        self.residual = residual

        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.atom_encoder = AtomEncoder(emb_dim)

        ### set the initial virtual node embedding to 0.
        self.virtualnode_embedding = torch.nn.Embedding(1, emb_dim)
        torch.nn.init.constant_(self.virtualnode_embedding.weight.data, 0)

        ### List of GNNs
        self.convs = torch.nn.ModuleList()
        ### batch norms applied to node embeddings
        self.batch_norms = torch.nn.ModuleList()

        ### List of MLPs to transform virtual node at every layer
        self.mlp_virtualnode_list = torch.nn.ModuleList()

        for layer in range(num_layer):
            if gnn_type == "gin":
                self.convs.append(GINConv(emb_dim))
            elif gnn_type == "gcn":
                self.convs.append(GCNConv(emb_dim))
            else:
                raise ValueError("Undefined GNN type called {}".format(gnn_type))

            self.batch_norms.append(torch.nn.BatchNorm1d(emb_dim))

        for layer in range(num_layer - 1):
            self.mlp_virtualnode_list.append(
                torch.nn.Sequential(
                    torch.nn.Linear(emb_dim, 2 * emb_dim),
                    torch.nn.BatchNorm1d(2 * emb_dim),
                    torch.nn.ReLU(),
                    torch.nn.Linear(2 * emb_dim, emb_dim),
                    torch.nn.BatchNorm1d(emb_dim),
                    torch.nn.ReLU(),
                )
            )

    def forward(self, batched_data):

        x, edge_index, edge_attr, batch = (
            batched_data.x,
            batched_data.edge_index,
            batched_data.edge_attr,
            batched_data.batch,
        )

        ### virtual node embeddings for graphs
        virtualnode_embedding = self.virtualnode_embedding(
            torch.zeros(batch[-1].item() + 1).to(edge_index.dtype).to(edge_index.device)
        )

        h_list = [self.atom_encoder(x)]
        for layer in range(self.num_layer):
            ### add message from virtual nodes to graph nodes
            h_list[layer] = h_list[layer] + virtualnode_embedding[batch]

            ### Message passing among graph nodes
            h = self.convs[layer](h_list[layer], edge_index, edge_attr)

            h = self.batch_norms[layer](h)
            if layer == self.num_layer - 1:
                # remove relu for the last layer
                h = F.dropout(h, self.drop_ratio, training=self.training)
            else:
                h = F.dropout(F.relu(h), self.drop_ratio, training=self.training)

            if self.residual:
                h = h + h_list[layer]

            h_list.append(h)

            ### update the virtual nodes
            if layer < self.num_layer - 1:
                ### add message from graph nodes to virtual nodes
                virtualnode_embedding_temp = (
                    global_add_pool(h_list[layer], batch) + virtualnode_embedding
                )
                ### transform virtual nodes using MLP

                if self.residual:
                    virtualnode_embedding = virtualnode_embedding + F.dropout(
                        self.mlp_virtualnode_list[layer](virtualnode_embedding_temp),
                        self.drop_ratio,
                        training=self.training,
                    )
                else:
                    virtualnode_embedding = F.dropout(
                        self.mlp_virtualnode_list[layer](virtualnode_embedding_temp),
                        self.drop_ratio,
                        training=self.training,
                    )

        ### Different implementations of Jk-concat
        if self.JK == "last":
            node_representation = h_list[-1]
        elif self.JK == "sum":
            node_representation = 0
            for layer in range(self.num_layer + 1):
                node_representation += h_list[layer]

        return node_representation


class GNN(torch.nn.Module):
    def __init__(
        self,
        num_tasks,
        num_layer=5,
        emb_dim=300,
        gnn_type="gin",
        virtual_node=True,
        residual=False,
        drop_ratio=0.5,
        JK="last",
        graph_pooling="mean",
    ):
        """
        num_tasks (int): number of labels to be predicted
        virtual_node (bool): whether to add virtual node or not
        """

        super().__init__()

        self.num_layer = num_layer
        self.drop_ratio = drop_ratio
        self.JK = JK
        self.emb_dim = emb_dim
        self.num_tasks = num_tasks
        self.graph_pooling = graph_pooling

        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        ### GNN to generate node embeddings
        if virtual_node:
            self.gnn_node = GNN_node_Virtualnode(
                num_layer,
                emb_dim,
                JK=JK,
                drop_ratio=drop_ratio,
                residual=residual,
                gnn_type=gnn_type,
            )
        else:
            self.gnn_node = GNN_node(
                num_layer,
                emb_dim,
                JK=JK,
                drop_ratio=drop_ratio,
                residual=residual,
                gnn_type=gnn_type,
            )

        ### Pooling function to generate whole-graph embeddings
        if self.graph_pooling == "sum":
            self.pool = global_add_pool
        elif self.graph_pooling == "mean":
            self.pool = global_mean_pool
        elif self.graph_pooling == "max":
            self.pool = global_max_pool
        elif self.graph_pooling == "attention":
            self.pool = GlobalAttention(
                gate_nn=torch.nn.Sequential(
                    torch.nn.Linear(emb_dim, 2 * emb_dim),
                    torch.nn.BatchNorm1d(2 * emb_dim),
                    torch.nn.ReLU(),
                    torch.nn.Linear(2 * emb_dim, 1),
                )
            )
        elif self.graph_pooling == "set2set":
            self.pool = Set2Set(emb_dim, processing_steps=2)
        else:
            raise ValueError("Invalid graph pooling type.")

        if graph_pooling == "set2set":
            self.graph_pred_linear = torch.nn.Linear(2 * self.emb_dim, self.num_tasks)
        else:
            self.graph_pred_linear = torch.nn.Linear(self.emb_dim, self.num_tasks)

    def forward(self, batched_data):
        h_node = self.gnn_node(batched_data)

        h_graph = self.pool(h_node, batched_data.batch)

        return self.graph_pred_linear(h_graph)


if __name__ == "__main__":
    GNN(num_tasks=10)



""" Building blocks for graph neural networks.
"""

import torch
import torch.nn as nn
from pytorch_lightning import LightningModule

from .base_class import *
from .representations import BaseRepresentation, TorchGeometricGraph

if isinstance(LightningModule, MagicMock):
    lm_imports = (BaseClass,)
else:
    lm_imports = (LightningModule, BaseClass)


class BaseLightningModule(*lm_imports):

    """BaseLightningModule allows for the use of a Pytorch Lightning module as a BaseClass to be incorporated into the framework.

    Parameters:
        optim (str, optional): parameter describing what kind of optimizer to use. Defaults to "adam".
        input_dimensions (Tuple, optional): Tulpe describing the dimensions of the input data. Defaults to None.
    """

    def __init__(self, optim: str = "adam", input_dimensions: Tuple = None):
        super().__init__()
        self.optim = optim

    def set_task_type(self, task_type, pos_weight=torch.tensor([1])):
        """Sets the task type for the model.

        Parameters:
            task_type (str): the task type to set the model to.
            pos_weight (torch.tensor, optional): the weight to use for the positive class. Defaults to torch.tensor([1]).

        """
        if task_type == "classification":
            self.loss_fun = nn.BCEWithLogitsLoss(
                reduction="none", pos_weight=pos_weight
            )
        else:
            self.loss_fun = nn.MSELoss()

    def loss(self, y_pred, y_true):
        """Calculate the loss for the model.

        Parameters:
            y_pred (torch.tensor): the predictions for the model.
            y_true (torch.tensor): the true labels for the model.

        Returns:
            torch.tensor: the loss for the model."""
        if len(y_true.shape) == 1:
            y_true = torch.unsqueeze(y_true, 1)
        y_available = ~torch.isnan(y_true)
        loss = self.loss_fun(y_pred[y_available], y_true[y_available])
        loss = loss.mean()
        return loss

    def _save(self):
        return ""

    def _load(self, d):
        pass

    def forward(self, batch):
        return self.network(batch)

    ##########################
    # train, val, test, steps
    ##########################
    def training_step(self, batch, batch_idx):
        y_true = batch.y.float()
        y_pred = self.forward(batch)
        loss = self.loss(y_pred, y_true)
        return loss

    def validation_step(self, batch, batch_idx):
        y_true = batch.y.float()
        y_pred = self.forward(batch)
        loss = self.loss(y_pred, y_true)
        return loss

    def test_step(self, batch, batch_idx):
        return self.validation_step(batch, batch_idx)

    ##################
    # Optimizers
    ##################
    def configure_optimizers(self):
        if self.optim == "adamw":
            if hasattr(self, "lr"):
                return torch.optim.AdamW(self.parameters(), lr=self.lr)
            else:
                return torch.optim.AdamW(self.parameters())
        else:
            if hasattr(self, "lr"):
                return torch.optim.Adam(self.parameters(), lr=self.lr)
            else:
                return torch.optim.Adam(self.parameters())


class MultiSequential(nn.Sequential):

    """Helper class to allow for the use of nn.Sequential with multi-input torch.nn modules"""

    def forward(self, *input):
        """Forward pass of the module.

        Parameters:
            *input (torch.tensor): the input data.

        Returns:
            torch.tensor: the output data."""
        x = None
        for module in self._modules.values():
            if x is None:
                x = module(*input)
            else:
                x = module(x)
        del input
        return x


class AttentiveFP(BaseLightningModule):

    """AttentiveFP is a wrapper for the PyTorch Geometric interpretation of https://pubs.acs.org/doi/10.1021/acs.jmedchem.9b00959.

    Parameters:
        hidden_channels (int, optional): the number of hidden channels to use in the model. Defaults to 4.
        out_channels (int, optional): the number of output channels to use in the model. Defaults to 1.
        num_layers (int, optional): the number of layers to use in the model. Defaults to 1.
        num_timesteps (int, optional): the number of timesteps to use in the model. Defaults to 1.
        dropout (float, optional): the dropout rate to use in the model. Defaults to 0.
        skip_lin (bool, optional): the number of skip connections to use in the model. Defaults to True.
        layer_dims (List, optional): the dimensions to use for each layer in the model. Defaults to [512, 128].
        activation (str, optional): the activation function to use in the model. Defaults to "leakyrelu".
        optim (str, optional): the optimizer to use in the model. Defaults to "adamw"."""

    @log_arguments
    def __init__(
        self,
        hidden_channels=4,
        out_channels=1,
        num_layers=1,
        num_timesteps=1,
        dropout=0,
        skip_lin=True,
        layer_dims=[512, 128],
        activation="leakyrelu",
        optim="adamw",
        **kwargs
    ):
        super().__init__(optim=optim, **kwargs)
        self.hidden_channels = hidden_channels
        self.out_channels = out_channels
        self.num_layers = num_layers
        self.num_timesteps = num_timesteps
        self.dropout = dropout
        self.layer_dims = [out_channels] + layer_dims

        self.skip_lin = skip_lin
        if skip_lin:
            self.out_channels = 1
            return

        if activation == "leakyrelu":
            self.activation = nn.LeakyReLU
        elif activation == "tanh":
            self.activation = nn.Tanh
        elif activation == "elu":
            self.activation = nn.ELU

    def create(self, dimensions):
        """Create the model.

        Parameters:
            dimensions (list): the dimensions of the input data.

        """

        from torch_geometric.nn.models import AttentiveFP as AttentiveFP_

        if self.skip_lin:
            self.network = AttentiveFP_(
                dimensions[0],
                self.hidden_channels,
                self.out_channels,
                dimensions[1],
                self.num_layers,
                self.num_timesteps,
                dropout=self.dropout,
            )
        else:
            self.network = MultiSequential(
                AttentiveFP_(
                    dimensions[0],
                    self.hidden_channels,
                    self.out_channels,
                    dimensions[1],
                    self.num_layers,
                    self.num_timesteps,
                    dropout=self.dropout,
                ),
                *[
                    nn.Sequential(
                        nn.Linear(self.layer_dims[i], self.layer_dims[i + 1]),
                        self.activation(),
                    )
                    for i in range(len(self.layer_dims) - 1)
                ],
                nn.Linear(self.layer_dims[-1], 1)
            )

    def forward(self, data):
        if data.x.is_cuda:
            dtype = torch.cuda.FloatTensor
        else:
            dtype = torch.FloatTensor
        x = data.x.type(dtype)
        edge_index = data.edge_index
        edge_attr = data.edge_attr.type(dtype)
        batch = data.batch

        del dtype
        del data
        return self.network(x, edge_index, edge_attr, batch)


class BaseTorchGeometricModel(BaseModel):

    """BaseTorchGeometricModel is a base class for models in the PyTorch Geometric framework.

    Parameters:
        network (nn.Module): The network to be used for the model.
        representation (BaseRepresentation, optional): The representation to be used for the model.
            Note that the representation must be compatible with the network,
            so the default, TorchGeometricGraph() is highly reccomended
        epochs (int, optional): The number of epochs to train the model for.
        batch_size (int, optional): The batch size to use for training.
        lr (float, optional): The learning rate to use for training.
        auto_lr_find (bool, optional): Whether to automatically adjust the learning rate.
        pos_weight (str, optional): Strategy for weighting positives in classification
        preinitialized (bool, optional): Whether to the network is pre-initialized.
        log (bool, optional): Log arguments or not. Should only be true if it is not nested. Defaults to True.
    """

    @log_arguments
    def __init__(
        self,
        network: BaseLightningModule,
        representation: BaseRepresentation = TorchGeometricGraph(),
        epochs: int = 1,
        batch_size: int = 16,
        lr: float = 1e-4,
        auto_lr_find: bool = True,
        pos_weight: str = "balanced",
        preinitialized: bool = False,
        log: bool = True,
        **kwargs
    ):
        self.representation = representation
        self.preinitialized = preinitialized
        self.network = network
        self.epochs = epochs
        self.batch_size = batch_size
        self.pos_weight = pos_weight

        from pytorch_lightning import Trainer

        self.trainer = Trainer(
            accelerator="auto", 
            max_epochs=self.epochs, 
            auto_lr_find=auto_lr_find,
            num_sanity_val_steps=0
        )

        super().__init__(log=False, **kwargs)

    def preprocess(self, X, y, fit=False):
        if y is None:
            y = [None] * len(X)

        return self.representation.convert(X, ys=y)

    def _fit(self, X_train, y_train):
        if hasattr(self.network, "create"):
            self.network.create((self.representation.dimensions))
        self.network.train()

        values, counts = np.unique(y_train, return_counts=True)
        if len(values) == 2:
            if self.pos_weight == "balanced":
                self.network.set_task_type(
                    "classification",
                    pos_weight=torch.tensor(
                        [(len(y_train) - sum(y_train)) / sum(y_train)]
                    ),
                )
            else:
                self.network.set_task_type("classification")
        else:
            self.network.set_task_type("regression")

        from torch_geometric.data import DataLoader as PyGDataLoader

        dataloader = PyGDataLoader(X_train, batch_size=self.batch_size, shuffle=True, num_workers=oce.CONFIG["NUM_WORKERS"])

        self.trainer.fit(self.network, dataloader)

    def _predict(self, X):
        self.network.eval()
        
        from torch_geometric.data import DataLoader as PyGDataLoader

        dataloader = PyGDataLoader(X, batch_size=self.batch_size, num_workers=oce.CONFIG["NUM_WORKERS"])

        predictions = self.trainer.predict(self.network, dataloader)

        predictions = np.concatenate([x.cpu().numpy() for x in predictions])

        return predictions.flatten()

    def _save(self) -> str:
        d = super()._save()
        buffer = io.BytesIO()
        torch.save(self.network, buffer)
        d.update({"save": buffer.getvalue()})
        return d

    def _load(self, d):
        super()._load(d)
        import sys

        self.network = torch.load(io.BytesIO(d["save"]))


from collections import OrderedDict


class TLFromCheckpoint(BaseLightningModule):

    """TLFromCheckpoint is a base class for transfer-learning from an OlorenVec PyTorch-lightning checkpoint.

    Parameters:
        model_path (str, option): The path to the PyTorch-lightning checkpoint. Ise
            "default" to use a pretrained OlorenVec model.
        map_location (str, optional): The location to map the model to. Default is "cuda:0".
        num_tasks (int, optional): The number of tasks in the OlorenVec model
        dropout (float, optional): The dropout rate to use for the model. Default is 0.1.
        lr (float, optional): The learning rate to use for training. Default is 1e-4.
        optim (str, optional): The optimizer to use for training. Default is "adam".
    """

    @log_arguments
    def __init__(
        self,
        model_path,
        map_location: str = "cuda:0",
        num_tasks: int = 2048,
        dropout: float = 0.1,
        lr: float = 1e-4,
        optim: str = "adam",
        reset: bool = False,
    ):
        self.lr = lr
        super().__init__(optim=optim)

        if model_path == "default":
            path = download_public_file("saves/olorenvec.ckpt")
        else:
            path = model_path

        if not torch.cuda.is_available():
            map_location = torch.device("cpu")
            logging.warn("Overriding map_location to cpu as no GPUs are available.")

        state_dict = OrderedDict(
            [
                (k.replace("model.", ""), v)
                for k, v in torch.load(path, map_location=map_location)[
                    "state_dict"
                ].items()
            ]
        )

        from olorenchemengine.pyg.gcn import GNN

        self.A = GNN(
            gnn_type="gcn",
            num_tasks=num_tasks,
            num_layer=5,
            emb_dim=300,
            drop_ratio=0.5,
            virtual_node=False,
        )
        if not reset:
            self.A.load_state_dict(state_dict)
        self.B = nn.Sequential(
            OrderedDict(
                [
                    ("dense1", nn.Linear(num_tasks, 64)),
                    ("relu1", nn.LeakyReLU()),
                    ("dropout1", nn.Dropout(dropout)),
                    ("dense2", nn.Linear(64, 1)),
                ]
            )
        )
        self.network = nn.Sequential(OrderedDict([("A", self.A), ("B", self.B)]))



""" Machine learning algorithms for use with molecular vector representations and features from experimental data.
"""

import io
import json
from tempfile import NamedTemporaryFile

import joblib
import numpy as np
import pandas as pd
import scipy
from pandas.api.types import is_numeric_dtype
from rdkit import Chem
from sklearn.model_selection import RandomizedSearchCV

import olorenchemengine as oce
from olorenchemengine.representations import (
    BaseCompoundVecRepresentation,
    BaseVecRepresentation,
)

from .base_class import *

try:
    from pytorch_lightning import LightningModule

    lm_imports = (LightningModule, BaseClass)
except:
    lm_imports = (BaseClass,)

if isinstance(RandomizedSearchCV, MagicMock):
    rscvmodel_extends = (BaseEstimator,)
else:
    rscvmodel_extends = (
        RandomizedSearchCV,
        BaseEstimator,
    )


class RandomizedSearchCVModel(*rscvmodel_extends):

    """Wrapper class for RandomizedSearchCV"""

    def __init__(self, *args, **kwargs):
        self.obj = None
        super().__init__(*args, **kwargs)

    def fit(self, *args, **kwargs):
        super().fit(*args, **kwargs)
        self.obj = self.best_estimator_

    def predict(self, *args, **kwargs):
        if self.obj is None:
            raise ValueError("Model not fitted")
        return self.obj.predict(*args, **kwargs)


class RandomForestRegressor(BaseEstimator):

    """Wrapper for sklearn RandomForestRegressor"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.ensemble import RandomForestRegressor

        self.obj = RandomForestRegressor(*args, **kwargs)


class RandomForestClassifier(BaseEstimator):

    """Wrapper for sklearn RandomForestClassifier"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.ensemble import RandomForestClassifier

        self.obj = RandomForestClassifier(*args, **kwargs)

    def predict(self, X):
        return [x[1] for x in self.obj.predict_proba(X)]


class SVC(BaseEstimator):

    """Wrapper for sklearn SVC"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.pipeline import make_pipeline
        from sklearn.preprocessing import StandardScaler
        from sklearn.svm import SVC

        self.obj = make_pipeline(StandardScaler(), SVC(*args, **kwargs))


class SVR(BaseEstimator):

    """Wrapper for sklearn SVR"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.pipeline import make_pipeline
        from sklearn.preprocessing import StandardScaler
        from sklearn.svm import SVR

        self.obj = make_pipeline(StandardScaler(), SVR(*args, **kwargs))


class KBestLinearRegression(BaseEstimator):

    """Selects the K-best features and then does linear regression"""

    @log_arguments
    def __init__(self, k=1, *args, **lwargs):
        from sklearn.feature_selection import SelectKBest, mutual_info_regression
        from sklearn.linear_model import LinearRegression
        from sklearn.pipeline import Pipeline

        self.obj = Pipeline(
            [
                ("KBest", SelectKBest(mutual_info_regression, k=k)),
                ("LR", LinearRegression()),
            ]
        )


class LogisticRegression(BaseEstimator):

    """Wrapper for sklearn LogisticRegression"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.linear_model import LogisticRegression

        self.obj = LogisticRegression(*args, **kwargs)


class RandomForestModel(BaseSKLearnModel):
    """Random forest model

        Parameters:
            n_estimators (int): The number of trees in the forest.
            max_depth (int): The maximum depth of the tree.
            max_features (int): The number of features to consider when looking for the best split.
            bootstrap (bool): Whether bootstrap samples are used when building trees.
            criterion (str): The function to measure the quality of a split.
            class_weight (str): Dict or 'balanced', defaults to None.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        representation,
        max_features="log2",
        max_depth=None,
        criterion="entropy",
        class_weight=None,
        bootstrap=True,
        n_estimators=100,
        **kwargs
    ):
        if not class_weight is None:
            try:
                class_weight = {
                    int(k): float(v) for k, v in json.loads(class_weight).items()
                }
            except:
                class_weight = None
        regressor = RandomForestRegressor(
            n_estimators=n_estimators,
            max_features=max_features,
            min_samples_split=4,
            min_samples_leaf=4,
            max_depth=max_depth,
            n_jobs=-1,
        )
        classifier = RandomForestClassifier(
            max_features=max_features,
            min_samples_split=4,
            min_samples_leaf=4,
            n_estimators=n_estimators,
            n_jobs=-1,
            criterion=criterion,
            bootstrap=bootstrap,
            max_depth=max_depth,
            class_weight=class_weight,
        )
        super().__init__(representation, regressor, classifier, log=False, **kwargs)


class AutoRandomForestModel(BaseSKLearnModel, BaseObject):
    """RandomForestModel where parameters have automatically tuned hyperparameters

        Parameters:
            representation (str): The representation to use for the model.
            n_iter (int): The number of iterations to run the hyperparameter tuning.
            scoring (str): The scoring metric to use for the hyperparameter tuning.
            verbose (int): The verbosity level of the hyperparameter tuning.
            cv (int): The number of folds to use for the hyperparameter tuning.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.AutoRandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048))
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self, representation, n_iter=100, scoring=None, verbose=2, cv=5, **kwargs
    ):
        from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

        regressor = self.autofit(RandomForestRegressor(), n_iter, cv, scoring, verbose)
        classifier = self.autofit(
            RandomForestClassifier(), n_iter, cv, scoring, verbose
        )
        super().__init__(representation, regressor, classifier, log=False, **kwargs)

    def autofit(self, model, n_iter, cv, scoring, verbose):
        """Takes an XGBoost model and replaces its fit function with one that
        automatically tunes the model hyperparameters

        Parameters:
            model (sklearn model): The model to be tuned
            n_iter (int): Number of iterations to run the hyperparameter tuning
            cv (int): Number of folds to use for cross-validation
            scoring (str): Scoring metric to use for cross-validation
            verbose (int): Verbosity level

        Returns:
            model (sklearn model): The tuned model
        """

        params = {
            "n_estimators": [50, 100, 200, 500, 1000],
            "max_depth": [3, 4, 5, 6, 7],
            "min_samples_split": [2, 3, 4, 5, 6],
            "min_samples_leaf": [1, 2, 3, 4, 5],
            "max_features": ["sqrt", "log2", "auto"],
            "min_impurity_decrease": [0.0, 0.1, 0.2, 0.3, 0.4],
            "max_samples": [0.5, 0.75, 1.0],
        }

        return RandomizedSearchCVModel(
            estimator=model,
            param_distributions=params,
            n_iter=n_iter,
            cv=cv,
            scoring=scoring,
            verbose=verbose,
        )


class BaseMLPClassifier(BaseEstimator):

    """Wrapper for sklearn MLP"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.neural_network import MLPClassifier

        self.obj = MLPClassifier(*args, **kwargs)


class BaseMLPRegressor(BaseEstimator):

    """Wrapper for sklearn MLP"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.neural_network import MLPRegressor

        self.obj = MLPRegressor(*args, **kwargs)


class SklearnMLP(BaseSKLearnModel):
    """MLP Model based on sklearn implementation

        Parameters:
            representation (BaseVecRepresentation): The representation to use for the model.
            hidden_layer_sizes (list): The number of neurons in each hidden layer.
            activation (str): The activation function to use.
            solver (str): The solver to use.
            alpha (float): Learning rate.
            batch_size (int): The size of the minibatches for stochastic optimizers.
            learning_rate (str): The learning rate schedule.
            learning_rate_init (float): The initial learning rate for the solver.
            power_t (float): The exponent for inverse scaling learning rate.
            max_iter (int): Maximum number of iterations.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.SklearnMLP(representation = oce.MorganVecRepresentation(radius=2, nbits=2048))
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        representation,
        hidden_layer_sizes=[100],
        activation="relu",
        solver="adam",
        alpha=0.0001,
        batch_size=32,
        learning_rate="constant",
        learning_rate_init=0.001,
        power_t=0.5,
        max_iter=200,
        log=True,
        **kwargs
    ):

        regressor = BaseMLPRegressor(
            hidden_layer_sizes=hidden_layer_sizes,
            activation=activation,
            learning_rate=learning_rate,
            solver=solver,
            alpha=alpha,
            batch_size=batch_size,
            learning_rate_init=learning_rate_init,
            power_t=power_t,
            max_iter=max_iter,
        )
        classifier = BaseMLPClassifier(
            hidden_layer_sizes=hidden_layer_sizes,
            activation=activation,
            learning_rate=learning_rate,
            solver=solver,
            alpha=alpha,
            batch_size=batch_size,
            learning_rate_init=learning_rate_init,
            power_t=power_t,
            max_iter=max_iter,
        )
        super().__init__(representation, regressor, classifier, log=False, **kwargs)


import pytorch_lightning as pl
import torch
import torch.nn as nn


class plMLP(pl.LightningModule):
    def __init__(self, hidden_layer_sizes, norm_layer, activation_layer, setting):
        super().__init__()
        hidden_layer_sizes = hidden_layer_sizes + [1]
        layers = []
        for i in range(len(hidden_layer_sizes) - 1):
            layers.append(nn.Linear(hidden_layer_sizes[i], hidden_layer_sizes[i + 1]))
            if not activation_layer is None and i < len(hidden_layer_sizes) - 2:
                layers.append(activation_layer())
            if not norm_layer is None and i < len(hidden_layer_sizes) - 2:
                layers.append(norm_layer())
        self.layers = nn.Sequential(*layers)
        if setting == "classification":
            self.layers.append(nn.Sigmoid())
            self.loss = nn.CrossEntropyLoss()
        else:
            self.loss = nn.MSELoss()

    def forward(self, x):
        return self.layers(x)

    def training_step(self, batch, batch_idx):
        x, y = batch
        x = x.view(x.size(0), -1)
        y_hat = self.layers(x)
        loss = self.loss(y_hat, y)
        self.log("train_loss", loss)
        return loss

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-4)
        return optimizer


class TorchMLP(BaseModel):
    """MLP Model based on torch implementation

        Parameters:
            representation (BaseVecRepresentation): The representation to use for the model.
            hidden_layer_sizes (list): The number of neurons in each hidden layer.
            norm_layer (str): The normalization to use for a final normalization layer. Default None.
            activation_layer (str): The activation function to use for a final activation layer. Default None.
            dropout (float): The dropout rate to use for the model.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.TorchMLP(representation = oce.MorganVecRepresentation(radius=2, nbits=2048))
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        representation,
        hidden_layer_sizes=[100],
        norm_layer: str = None,
        activation_layer: str = None,
        dropout: float = 0.0,
        log=True,
        **kwargs
    ):
        self.representation = representation
        self.hidden_layer_sizes = hidden_layer_sizes
        import torch.nn as nn

        if activation_layer == "leakyrelu":
            self.activation_layer = nn.LeakyReLU
        elif activation_layer == "relu":
            self.activation_layer = nn.ReLU
        elif activation_layer == "tanh":
            self.activation_layer = nn.Tanh
        elif activation_layer == "sigmoid":
            self.activation_layer = nn.Sigmoid
        else:
            self.activation_layer = None

        if norm_layer == "batchnorm":
            self.norm_layer = nn.BatchNorm1d
        elif norm_layer == "layernorm":
            self.norm_layer = nn.LayerNorm
        else:
            self.norm_layer = None

        self.dropout = dropout

        super().__init__(representation, log=False)

    def preprocess(self, X, y, fit=False):
        if self.representation is None:
            return X
        else:
            return self.representation.convert(X, fit=fit)

    def _fit(self, X, y):
        import pytorch_lightning as pl
        import torch
        import torch.nn as nn

        self.network = plMLP(
            hidden_layer_sizes=[len(X[0])] + self.hidden_layer_sizes,
            norm_layer=self.norm_layer,
            activation_layer=self.activation_layer,
            setting=self.setting,
        )

        from torch.utils.data import DataLoader, TensorDataset

        y = np.array(y.tolist())
        device = torch.device(oce.CONFIG["MAP_LOCATION"])
        dataset = TensorDataset(
            torch.from_numpy(X).float().to(device),
            torch.from_numpy(y).float().to(device),
        )
        loader = DataLoader(dataset, batch_size=32)

        self.trainer = pl.Trainer(
            accelerator="auto",
            devices=1
            if torch.cuda.is_available()
            else None,  # limiting got iPython runs
            max_epochs=10,
        )

        self.trainer.fit(self.network, loader)

    def _predict(self, X):
        import torch

        # from torch.utils.data import TensorDataset, DataLoader
        device = torch.device(oce.CONFIG["MAP_LOCATION"])
        # dataset = TensorDataset(torch.from_numpy(X).float().to(device))
        # loader = DataLoader(dataset, batch_size=1)
        # return self.trainer.predict(self.network, loader)
        self.network.to(device)
        return (
            self.network(torch.from_numpy(X).float().to(device)).detach().cpu().numpy()
        )

    def _save(self) -> str:
        d = super()._save()
        if not hasattr(self, "network"):
            return d
        buffer = io.BytesIO()
        import torch

        torch.save(self.network, buffer)
        d.update({"save": buffer.getvalue()})
        return d

    def _load(self, d):
        super()._load(d)
        if not "save" in d:
            return
        import torch

        self.network = torch.load(io.BytesIO(d["save"]))


class XGBoostModel(BaseSKLearnModel, BaseObject):
    """XGBoost model

        Parameters:
            representation (str): The representation to use for the model
            n_iter (int): Number of iterations to run the hyperparameter tuning
            cv (int): Number of folds to use for cross-validation
            scoring (str): Scoring metric to use for cross-validation
            verbose (int): Verbosity level

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.XGBoostModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048))
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        representation,
        n_estimators=2000,
        max_depth=7,
        subsample=0.5,
        max_leaves=5,
        learning_rate=0.05,
        colsample_bytree=0.8,
        **kwargs
    ):
        from xgboost import XGBClassifier, XGBRegressor

        regressor = XGBRegressor(
            tree_method="gpu_hist",
            n_estimators=n_estimators,
            max_depth=max_depth,
            max_leaves=max_leaves,
            learning_rate=learning_rate,
            subsample=subsample,
            colsample_bytree=colsample_bytree,
        )

        classifier = XGBClassifier(
            tree_method="gpu_hist",
            n_estimators=n_estimators,
            max_depth=max_depth,
            max_leaves=max_leaves,
            learning_rate=learning_rate,
            subsample=subsample,
            colsample_bytree=colsample_bytree,
        )

        super().__init__(representation, regressor, classifier, log=False, **kwargs)


class ZWK_XGBoostModel(BaseSKLearnModel, BaseObject):
    """XGBoost model from https://github.com/smu-tao-group/ADMET_XGBoost

        Parameters:
            representation (str): The representation to use for the model.
            n_iter (int): The number of iterations to run the hyperparameter tuning.
            scoring (str): The scoring metric to use for the hyperparameter tuning.
            verbose (int): The verbosity level of the hyperparameter tuning.
            cv (int): The number of folds to use for the hyperparameter tuning.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.ZWK_XGBoostModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048))
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self, representation, n_iter=100, scoring=None, verbose=2, cv=5, **kwargs
    ):
        from xgboost import XGBClassifier, XGBRegressor

        if scoring == "spearman":
            from scipy.stats import spearmanr
            from sklearn.metrics import make_scorer

            def spearman_loss_func(y_true, y_pred):
                """spearman metric"""
                return spearmanr(y_true, y_pred)[0]

            spearman = make_scorer(spearman_loss_func)
            scoring = spearman

        regressor = self.autofit(
            XGBRegressor(tree_method="gpu_hist"), n_iter, cv, scoring, verbose
        )
        classifier = self.autofit(
            XGBClassifier(tree_method="gpu_hist"), n_iter, cv, scoring, verbose
        )

        super().__init__(representation, regressor, classifier, log=False, **kwargs)

    def autofit(self, model, n_iter, cv, scoring, verbose):
        """Takes an XGBoost model and replaces its fit function with one that
        automatically tunes the model hyperparameters

        Parameters:
            model (sklearn model): The model to be tuned
            n_iter (int): Number of iterations to run the hyperparameter tuning
            cv (int): Number of folds to use for cross-validation
            scoring (str): Scoring metric to use for cross-validation
            verbose (int): Verbosity level

        Returns:
            model (sklearn model): The tuned model
        """

        params = {
            "n_estimators": [50, 100, 200, 500, 1000],
            "max_depth": [3, 4, 5, 6, 7],
            "learning_rate": [0.001, 0.005, 0.01, 0.1],
            "subsample": [0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            "colsample_bytree": [0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            "reg_alpha": [0, 0.1, 1, 5, 10, 20, 25, 30],
            "reg_lambda": [0, 0.1, 1, 5, 10, 20, 25, 30],
            "min_child_weight": [1, 2, 3],
        }

        return RandomizedSearchCVModel(
            estimator=model,
            param_distributions=params,
            n_iter=n_iter,
            cv=cv,
            scoring=scoring,
            verbose=verbose,
        )


class SupportVectorMachine(BaseSKLearnModel):

    """Support vector machine

        Parameters:
            representations (str): The representation to use for the model.
            kernel (str): The kernel to use for the model.
            C (float): The C parameter for the model.
            gamma (float): The gamma parameter for the model.
            coef0 (float): The coef0 parameter for the model.
            cache_size (int): The cache size parameter for the model.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.SupportVectorMachine(representation = oce.MorganVecRepresentation(radius=2, nbits=2048))
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        representation,
        C=0.8,
        kernel="rbf",
        gamma="scale",
        coef0=0,
        cache_size=500,
        **kwargs
    ):
        regressor = SVR(
            C=C,
            kernel=kernel,
            gamma=gamma,
            coef0=coef0,
            cache_size=cache_size,
        )
        classifier = SVC(
            C=C,
            kernel=kernel,
            gamma=gamma,
            coef0=coef0,
            cache_size=cache_size,
        )
        super().__init__(representation, regressor, classifier, log=False, **kwargs)


class KNeighborsRegressor(BaseEstimator):

    """Wrapper for sklearn KNeighborsRegressor"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.neighbors import KNeighborsRegressor

        self.obj = KNeighborsRegressor(*args, **kwargs)


class KNeighborsClassifier(BaseEstimator):

    """Wrapper for sklearn KNeighborsClassifier"""

    @log_arguments
    def __init__(self, *args, **kwargs):
        from sklearn.neighbors import KNeighborsClassifier

        self.obj = KNeighborsClassifier(*args, **kwargs)

    def predict(self, X):
        return [x[1] for x in self.obj.predict_proba(X)]


class KNN(BaseSKLearnModel):

    """KNN model

        Parameters:
            representations (str): The representation to use for the model.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.KNN(representation = oce.MorganVecRepresentation(radius=2, nbits=2048))
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(self, representation, **kwargs):
        regressor = KNeighborsRegressor(**kwargs)
        classifier = KNeighborsClassifier(**kwargs)
        super().__init__(representation, regressor, classifier, log=False, **kwargs)


class MLP(BaseSKLearnModel):

    """MLP model

        Parameters:
            representation (BaseVecRepresentation): The representation to use for the model.
            hidden_layer_sizes (List[int]): The hidden layer sizes to use for the model.
            activation (str): The activation function to use for the model.
            epochs (int): The number of epochs to use for the model.
            batch_size (int): The batch size to use for the model.
            lr (float): The learning rate to use for the model.
            dropout (float): The dropout rate to use for the model.
            kernel_regularizer (float): The kernel regularizer to use for the model.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.MLP(representation = oce.MorganVecRepresentation(radius=2, nbits=2048))
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    def preprocess(self, X, y, fit=False):
        if self.representation is None:
            return np.array(X)
        else:
            return np.array(self.representation.convert(X, fit=fit))

    # specify parameters
    @log_arguments
    def __init__(
        self,
        representation: BaseVecRepresentation,
        layer_dims=[2048, 512, 128],
        activation="tanh",
        epochs=100,
        batch_size=16,
        lr=0.0005,
        dropout=0,
        kernel_regularizer=1e-4,
        **kwargs
    ):

        super().__init__(
            representation,
            TorchMLP(
                None,
                layer_dims=layer_dims,
                activation=activation,
                epochs=epochs,
                batch_size=batch_size,
                dropout=dropout,
                kernel_regularizer=kernel_regularizer,
                lr=lr,
                task="regression",
            ),
            TorchMLP(
                None,
                layer_dims=layer_dims,
                activation=activation,
                epochs=epochs,
                batch_size=batch_size,
                dropout=dropout,
                kernel_regularizer=kernel_regularizer,
                lr=lr,
                task="classification",
            ),
            log=False,
            **kwargs
        )


class GuessingRegression(BaseModel):

    """Guessing model for regression, used to infer non-linear relationships.

    This model tries different non-linear relationships between each feature and
    property, selecting the best such relationship for each feature. Then the
    features are transformed then either aggregated using linear regression or
    averages to obtain the final prediction for the property. This is best used
    for using experimental features with direct relationships to the properties.

    Attributes:
        transformations (List[Callable]): List of transformations to apply to
            the data i.e. possible relationships between feature and property.
        state (Dict[str, Tuple[int, float, float]]): State of the model,
            best transformation for each feature.
        reg (str): Method to use for combining features,
            either "lr" linear regression or "avg" average."""

    @log_arguments
    def __init__(self, config="full", reg="lr", **kwargs):

        safe_log = lambda x: np.log10(np.abs(x + 1e-3))
        safe_divide = lambda x1, x2: x1 / (np.abs(x2) + 1e-3)

        self.transformations = [
            (
                lambda x: safe_log(x),
                lambda y: safe_log(y),
                lambda A, B: lambda x: np.exp(B + A * np.log(x)),
                "x: log x, y: log y",
            ),
            (lambda x: x, lambda y: y, lambda A, B: lambda x: A * x + B, "x: x, y: y"),
            (
                lambda x: x,
                lambda y: safe_log(y),
                lambda A, B: lambda x: np.exp(B + A * x),
                "x: x, y: log y",
            ),
            (
                lambda x: safe_log(x),
                lambda y: y,
                lambda A, B: lambda x: B * safe_log(x) + A,
                "x: log x, y: y",
            ),
            (
                lambda x: safe_divide(1, x),
                lambda y: y,
                lambda A, B: lambda x: safe_divide((A + B * x), x),
                "x: 1/x, y: y",
            ),
            (
                lambda x: x,
                lambda y: safe_divide(1, y),
                lambda A, B: lambda x: safe_divide(1, (B + A * x)),
                "x: x, y: 1/y",
            ),
            (
                lambda x: safe_divide(1, x),
                lambda y: safe_divide(1, y),
                lambda A, B: lambda x: safe_divide(1, (B + A * x)),
                "x: 1/x, y: 1/y",
            ),
        ]
        self.state = {}
        if reg == "lr":
            self.reg = LinearRegression()
        if reg == "avg":
            self.reg = None
        super().__init__(log=False, **kwargs)

    def preprocess(self, X, y, fit=False):
        """This method is used to preprocess the data before training.

        Parameters:
            X (np.ndarray): List of lists of features.
            y (np.array): List of properties.

        Returns:
            np.ndarray: List of lists of features.
            Shape: (n_samples, n_features)"""
        return X.copy()

    def linearize(self, X):
        """Linearize the data--apply the best transformation for each feature.

        Parameters:
            X (np.ndarray): List of lists of features.

        Returns:
            np.ndarray: List of lists of features.
            Shape: (n_samples, n_features)"""
        out = pd.DataFrame()
        for c in self.state.keys():
            trans = self.transformations[self.state[c][0]][2]
            out[c] = trans(self.state[c][1], self.state[c][2])(X[c])
        return out

    def _fit(self, X_train, y_train):
        assert isinstance(X_train, pd.DataFrame)
        best = ""
        for c in X_train.columns.tolist():
            if is_numeric_dtype(X_train[c]):
                df2 = pd.DataFrame({"x": X_train[c], "y": y_train}).dropna()
                r2max = -1
                for i, trans in enumerate(self.transformations):
                    t0 = trans[0](df2["x"])
                    t1 = trans[1](df2["y"])
                    A, B, r, p, se = scipy.stats.linregress(t0, t1)
                    if r**2 > r2max:
                        r2max = r**2
                        best = trans[3]
                        self.state[c] = (i, A, B, r2max)
                print(c)
                print(best)
                print(r2max)

        X_train = self.linearize(X_train)
        X_train["y"] = y_train
        X_train = X_train.dropna()
        if not self.reg is None:
            self.reg.fit(X_train.loc[:, X_train.columns != "y"], X_train["y"])

    def _predict(self, X):
        X = self.linearize(X)
        if self.reg == "lr":
            return self.reg.predict(X)
        else:
            return X.mean(axis=1)

    def _save(self):
        d = super()._save()
        if not self.reg is None:
            d.update({"reg": self.reg._save(), "state": self.state})
        else:
            d.update({"state": self.state})
        return d

    def _load(self, d):
        super()._load(d)
        if not self.reg is None:
            self.reg._load(d["reg"])
        self.state = d["state"]


class FeaturesClassification(BaseModel):

    """FeaturesClassification uses machine learning models to classify features
    based on their experimental data

    Attributes:
        obj (sklearn.base.BaseEstimator): Machine learning model to use.

    Parameters:
        config (str): Configuration to use for the model."""

    @log_arguments
    def __init__(self, config="lineardiscriminant"):
        """FeaturesClassfication constructor

        Args:
            config (str, optional): Type of machine learning model to use to
                classify features. Options are "lineardiscriminant",
                "gaussiannb", "quadraticdiscriminant", "knearestneighbors".
                Defaults to "lineardiscriminant".
        """
        self.numeric_cols = []
        if config == "lineardiscriminant":
            from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

            self.obj = LinearDiscriminantAnalysis()
        elif config == "gaussiannb":
            from sklearn.naive_bayes import GaussianNB

            self.obj = GaussianNB()
        elif config == "quadraticdiscriminant":
            from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis

            self.obj = QuadraticDiscriminantAnalysis()
        elif config == "knearestneighbors":
            from sklearn.neighbors import KNeighborsClassifier

            self.obj = KNeighborsClassifier()

    def _fit(self, X_train, y_train):
        self.numeric_cols = [
            c for c in X_train.columns.tolist() if is_numeric_dtype(X_train[c])
        ]
        X_train = X_train[self.numeric_cols]
        self.obj.fit(X_train, y_train)

    def _predict(self, X):
        X = X[self.numeric_cols]
        return self.obj.predict(X).astype(float)

    def _save(self):
        b = io.BytesIO()
        joblib.dump(self.obj, b)
        return {"obj": b.getvalue(), "numeric_cols": self.numeric_cols}

    def _load(self, d):
        super()._load(d)
        self.obj = joblib.load(io.BytesIO(d["obj"]))
        self.numeric_cols = d["numeric_cols"]



def _is_object(key):
    return key.split(".")[-1][0].isupper()


class RemoteSession:
    def __init__(self):
        pass

    def construct(self, object):
        if issubclass(type(object), RemoteObject):
            if hasattr(object, "REMOTE_ID"):  # object already constructed
                return object
            if _is_object(object.REMOTE_BC_KEY):
                object.REMOTE_ID = generate_random_uuid()
                print(f"CONSTRUCT {object.REMOTE_ID} => {parameterize_remote(object)}")
                return object
            else:
                return self.get_attribute(
                    object.REMOTE_PARENT, object.REMOTE_BC_KEY.split(".")[-1]
                )
        elif (
            object is None
            or issubclass(type(object), int)
            or issubclass(type(object), float)
            or issubclass(type(object), str)
        ):
            return object
        elif issubclass(type(object), list):
            return [self.construct(x) for x in object]

    def run_method(self, object, method_name, args, kwargs):
        object = self.construct(object)
        args = [self.construct(arg) for arg in args]
        kwargs = {k: self.construct(v) for k, v in kwargs.items()}
        print(
            f"RUN {object.REMOTE_ID}.{method_name} with args = {args}, and kwargs = {kwargs}"
        )

    def get_attribute(self, object, attribute_name):
        object = self.construct(object)
        object.attribute_name.REMOTE_ID = generate_random_uuid()
        print(
            f"GET_ATTR {object.REMOTE_ID}.{attribute_name} => {object.attribute_name.REMOTE_ID}"
        )
        return object.attribute_name.REMOTE_ID

    def __getattribute__(self, key):
        EXCLUDED_WORDS = ["run_method", "construct", "get_attribute"]
        if key in EXCLUDED_WORDS:
            return object.__getattribute__(self, key)
        return RemoteObject(key, None, self)


def generate_random_uuid():
    import uuid

    return str(uuid.uuid4())


def parameterize_remote(object):
    if issubclass(type(object), RemoteObject):
        if not hasattr(object, "args") or not hasattr(object, "kwargs"):
            print("Warning: RemoteObject has no args or kwargs")
            print(f"Object is {object.REMOTE_BC_KEY}")
        return {
            **(
                {"BC_remote_id": object.REMOTE_ID}
                if hasattr(object, "REMOTE_ID")
                else {}
            ),
            **{"BC_class_name": object.REMOTE_BC_KEY.split(".")[-1]},
            **{"args": [parameterize_remote(arg) for arg in object.args]},
            **{"kwargs": {k: parameterize_remote(v) for k, v in object.kwargs.items()}},
        }
    elif (
        object is None
        or issubclass(type(object), int)
        or issubclass(type(object), float)
        or issubclass(type(object), str)
    ):
        return object
    elif issubclass(type(object), list):
        return [parameterize_remote(x) for x in object]
    else:
        raise ValueError(f"Cannot parameterize {object} of type {type(object)}")


class RemoteObject:
    def __init__(self, key, parent) -> None:
        self.REMOTE_BC_KEY = key
        self.REMOTE_PARENT = parent
        self.REMOTE_CHILDREN = {}

    def __getattribute__(self, key):
        EXCLUDED_WORDS = [
            "REMOTE_BC_KEY",
            "REMOTE_SESSION",
            "REMOTE_PARENT",
            "REMOTE_ID",
            "REMOTE_CHILDREN",
            "args",
            "kwargs",
            "shabalaba",
        ]

        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            if key in self.REMOTE_CHILDREN:
                return self.REMOTE_CHILDREN[key]
            else:
                self.REMOTE_CHILDREN[key] = RemoteObject(
                    self.REMOTE_BC_KEY + "." + key, self, self.REMOTE_SESSION
                )
                return self.REMOTE_CHILDREN[key]

    def __repr__(self) -> str:
        if hasattr(self, "args") or hasattr(self, "kwargs"):
            return parameterize_remote(self)
        return "RemoteObject(" + self.REMOTE_BC_KEY + ")"

    def __call__(self, *args, **kwargs):
        if self.REMOTE_BC_KEY.split(".")[-1][0].isupper():  # instantiating an object
            self.args = args
            self.kwargs = kwargs
            return self
        else:
            return self.REMOTE_SESSION.run_method(
                self.REMOTE_PARENT, self.REMOTE_BC_KEY.split(".")[-1], args, kwargs
            )
            # parameterized_kwargs = {parameterize_remote(k): parameterize_remote(v) for k, v in kwargs.items()}
            # print(f"Calling function {self.REMOTE_BC_KEY.split('.')[-1]} with: args: {[parameterize_remote(a) for a in args]}, kwargs: {parameterized_kwargs} on {parameterize_remote(self.REMOTE_PARENT)}" )
            # return self # TODO: handle calling a function (like train)

    # def __iter__(self):
    #     return iter([0, 1])


# TODO: track operators on RemoteObjects

# We are basically making a Python "Compiler" that compiles down to a JSON dictionary - but we can use python
# in the compiler lol


if __name__ == "__main__":
    # oce = RemoteSession()
    # dataset = oce.ExampleDataset()
    # dataset.shabalaba

    import olorenchemengine as oce

    print(dir(oce))

    # sub_model1 = oce.RandomForestModel(oce.DescriptastorusDescriptor("rdkit2dnormalized"), n_estimators=1000)
    # model = oce.ensemble.BaseBoosting([
    #         sub_model1,
    #         oce.RandomForestModel(oce.OlorenCheckpoint("default"), n_estimators=1000)])
    # model.fit(dataset.train_dataset)
    # oce.save(model, "model.oce")

    # CONSTRUCT model FROM JSON_PARAMS
    # CONSTRUCT dataset FROM JSON_PARAMS
    # CONSTRUCT train_dataset FROM dataset.train_dataset
    # RUN fit on model with train_dataset



"""
colors

The MIT License (MIT)

Copyright (c) 2016-2018 Plotly, Inc

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

=====

Functions that manipulate colors and arrays of colors.

-----
There are three basic types of color types: rgb, hex and tuple:

rgb - An rgb color is a string of the form 'rgb(a,b,c)' where a, b and c are
integers between 0 and 255 inclusive.

hex - A hex color is a string of the form '#xxxxxx' where each x is a
character that belongs to the set [0,1,2,3,4,5,6,7,8,9,a,b,c,d,e,f]. This is
just the set of characters used in the hexadecimal numeric system.

tuple - A tuple color is a 3-tuple of the form (a,b,c) where a, b and c are
floats between 0 and 1 inclusive.

-----
Colormaps and Colorscales:
A colormap or a colorscale is a correspondence between values - Pythonic
objects such as strings and floats - to colors.

There are typically two main types of colormaps that exist: numerical and
categorical colormaps.

Numerical:
----------
Numerical colormaps are used when the coloring column being used takes a
spectrum of values or numbers.

A classic example from the Plotly library:
```
rainbow_colorscale =  [
    [0, 'rgb(150,0,90)'], [0.125, 'rgb(0,0,200)'],
    [0.25, 'rgb(0,25,255)'], [0.375, 'rgb(0,152,255)'],
    [0.5, 'rgb(44,255,150)'], [0.625, 'rgb(151,255,0)'],
    [0.75, 'rgb(255,234,0)'], [0.875, 'rgb(255,111,0)'],
    [1, 'rgb(255,0,0)']
]
```

Notice that this colorscale is a list of lists with each inner list containing
a number and a color. These left hand numbers in the nested lists go from 0 to
1, and they are like pointers tell you when a number is mapped to a specific
color.

If you have a column of numbers `col_num` that you want to plot, and you know

```
min(col_num) = 0
max(col_num) = 100
```

then if you pull out the number `12.5` in the list and want to figure out what
color the corresponding chart element (bar, scatter plot, etc) is going to be,
you'll figure out that proportionally 12.5 to 100 is the same as 0.125 to 1.
So, the point will be mapped to 'rgb(0,0,200)'.

All other colors between the pinned values in a colorscale are linearly
interpolated.

Categorical:
------------
Alternatively, a categorical colormap is used to assign a specific value in a
color column to a specific color everytime it appears in the dataset.

A column of strings in a panadas.dataframe that is chosen to serve as the
color index would naturally use a categorical colormap. However, you can
choose to use a categorical colormap with a column of numbers.

Be careful! If you have a lot of unique numbers in your color column you will
end up with a colormap that is massive and may slow down graphing performance.
"""
from __future__ import absolute_import

import decimal
from numbers import Number

DEFAULT_PLOTLY_COLORS = [
    "rgb(31, 119, 180)",
    "rgb(255, 127, 14)",
    "rgb(44, 160, 44)",
    "rgb(214, 39, 40)",
    "rgb(148, 103, 189)",
    "rgb(140, 86, 75)",
    "rgb(227, 119, 194)",
    "rgb(127, 127, 127)",
    "rgb(188, 189, 34)",
    "rgb(23, 190, 207)",
]

PLOTLY_SCALES = {
    "Greys": [[0, "rgb(0,0,0)"], [1, "rgb(255,255,255)"]],
    "YlGnBu": [
        [0, "rgb(8,29,88)"],
        [0.125, "rgb(37,52,148)"],
        [0.25, "rgb(34,94,168)"],
        [0.375, "rgb(29,145,192)"],
        [0.5, "rgb(65,182,196)"],
        [0.625, "rgb(127,205,187)"],
        [0.75, "rgb(199,233,180)"],
        [0.875, "rgb(237,248,217)"],
        [1, "rgb(255,255,217)"],
    ],
    "Greens": [
        [0, "rgb(0,68,27)"],
        [0.125, "rgb(0,109,44)"],
        [0.25, "rgb(35,139,69)"],
        [0.375, "rgb(65,171,93)"],
        [0.5, "rgb(116,196,118)"],
        [0.625, "rgb(161,217,155)"],
        [0.75, "rgb(199,233,192)"],
        [0.875, "rgb(229,245,224)"],
        [1, "rgb(247,252,245)"],
    ],
    "YlOrRd": [
        [0, "rgb(128,0,38)"],
        [0.125, "rgb(189,0,38)"],
        [0.25, "rgb(227,26,28)"],
        [0.375, "rgb(252,78,42)"],
        [0.5, "rgb(253,141,60)"],
        [0.625, "rgb(254,178,76)"],
        [0.75, "rgb(254,217,118)"],
        [0.875, "rgb(255,237,160)"],
        [1, "rgb(255,255,204)"],
    ],
    "Bluered": [[0, "rgb(0,0,255)"], [1, "rgb(255,0,0)"]],
    # modified RdBu based on
    # www.sandia.gov/~kmorel/documents/ColorMaps/ColorMapsExpanded.pdf
    "RdBu": [
        [0, "rgb(5,10,172)"],
        [0.35, "rgb(106,137,247)"],
        [0.5, "rgb(190,190,190)"],
        [0.6, "rgb(220,170,132)"],
        [0.7, "rgb(230,145,90)"],
        [1, "rgb(178,10,28)"],
    ],
    # Scale for non-negative numeric values
    "Reds": [
        [0, "rgb(220,220,220)"],
        [0.2, "rgb(245,195,157)"],
        [0.4, "rgb(245,160,105)"],
        [1, "rgb(178,10,28)"],
    ],
    # Scale for non-positive numeric values
    "Blues": [
        [0, "rgb(5,10,172)"],
        [0.35, "rgb(40,60,190)"],
        [0.5, "rgb(70,100,245)"],
        [0.6, "rgb(90,120,245)"],
        [0.7, "rgb(106,137,247)"],
        [1, "rgb(220,220,220)"],
    ],
    "Picnic": [
        [0, "rgb(0,0,255)"],
        [0.1, "rgb(51,153,255)"],
        [0.2, "rgb(102,204,255)"],
        [0.3, "rgb(153,204,255)"],
        [0.4, "rgb(204,204,255)"],
        [0.5, "rgb(255,255,255)"],
        [0.6, "rgb(255,204,255)"],
        [0.7, "rgb(255,153,255)"],
        [0.8, "rgb(255,102,204)"],
        [0.9, "rgb(255,102,102)"],
        [1, "rgb(255,0,0)"],
    ],
    "Rainbow": [
        [0, "rgb(150,0,90)"],
        [0.125, "rgb(0,0,200)"],
        [0.25, "rgb(0,25,255)"],
        [0.375, "rgb(0,152,255)"],
        [0.5, "rgb(44,255,150)"],
        [0.625, "rgb(151,255,0)"],
        [0.75, "rgb(255,234,0)"],
        [0.875, "rgb(255,111,0)"],
        [1, "rgb(255,0,0)"],
    ],
    "Portland": [
        [0, "rgb(12,51,131)"],
        [0.25, "rgb(10,136,186)"],
        [0.5, "rgb(242,211,56)"],
        [0.75, "rgb(242,143,56)"],
        [1, "rgb(217,30,30)"],
    ],
    "Jet": [
        [0, "rgb(0,0,131)"],
        [0.125, "rgb(0,60,170)"],
        [0.375, "rgb(5,255,255)"],
        [0.625, "rgb(255,255,0)"],
        [0.875, "rgb(250,0,0)"],
        [1, "rgb(128,0,0)"],
    ],
    "Hot": [
        [0, "rgb(0,0,0)"],
        [0.3, "rgb(230,0,0)"],
        [0.6, "rgb(255,210,0)"],
        [1, "rgb(255,255,255)"],
    ],
    "Blackbody": [
        [0, "rgb(0,0,0)"],
        [0.2, "rgb(230,0,0)"],
        [0.4, "rgb(230,210,0)"],
        [0.7, "rgb(255,255,255)"],
        [1, "rgb(160,200,255)"],
    ],
    "Earth": [
        [0, "rgb(0,0,130)"],
        [0.1, "rgb(0,180,180)"],
        [0.2, "rgb(40,210,40)"],
        [0.4, "rgb(230,230,50)"],
        [0.6, "rgb(120,70,20)"],
        [1, "rgb(255,255,255)"],
    ],
    "Electric": [
        [0, "rgb(0,0,0)"],
        [0.15, "rgb(30,0,100)"],
        [0.4, "rgb(120,0,100)"],
        [0.6, "rgb(160,90,0)"],
        [0.8, "rgb(230,200,0)"],
        [1, "rgb(255,250,220)"],
    ],
    "Viridis": [
        [0, "#440154"],
        [0.06274509803921569, "#48186a"],
        [0.12549019607843137, "#472d7b"],
        [0.18823529411764706, "#424086"],
        [0.25098039215686274, "#3b528b"],
        [0.3137254901960784, "#33638d"],
        [0.3764705882352941, "#2c728e"],
        [0.4392156862745098, "#26828e"],
        [0.5019607843137255, "#21918c"],
        [0.5647058823529412, "#1fa088"],
        [0.6274509803921569, "#28ae80"],
        [0.6901960784313725, "#3fbc73"],
        [0.7529411764705882, "#5ec962"],
        [0.8156862745098039, "#84d44b"],
        [0.8784313725490196, "#addc30"],
        [0.9411764705882353, "#d8e219"],
        [1, "#fde725"],
    ],
    "Cividis": [
        [0.000000, "rgb(0,32,76)"],
        [0.058824, "rgb(0,42,102)"],
        [0.117647, "rgb(0,52,110)"],
        [0.176471, "rgb(39,63,108)"],
        [0.235294, "rgb(60,74,107)"],
        [0.294118, "rgb(76,85,107)"],
        [0.352941, "rgb(91,95,109)"],
        [0.411765, "rgb(104,106,112)"],
        [0.470588, "rgb(117,117,117)"],
        [0.529412, "rgb(131,129,120)"],
        [0.588235, "rgb(146,140,120)"],
        [0.647059, "rgb(161,152,118)"],
        [0.705882, "rgb(176,165,114)"],
        [0.764706, "rgb(192,177,109)"],
        [0.823529, "rgb(209,191,102)"],
        [0.882353, "rgb(225,204,92)"],
        [0.941176, "rgb(243,219,79)"],
        [1.000000, "rgb(255,233,69)"],
    ],
}


def color_parser(colors, function):
    """
    Takes color(s) and a function and applies the function on the color(s)

    In particular, this function identifies whether the given color object
    is an iterable or not and applies the given color-parsing function to
    the color or iterable of colors. If given an iterable, it will only be
    able to work with it if all items in the iterable are of the same type
    - rgb string, hex string or tuple
    """
    if isinstance(colors, str):
        return function(colors)

    if isinstance(colors, tuple) and isinstance(colors[0], Number):
        return function(colors)

    if hasattr(colors, "__iter__"):
        if isinstance(colors, tuple):
            new_color_tuple = tuple(function(item) for item in colors)
            return new_color_tuple

        else:
            new_color_list = [function(item) for item in colors]
            return new_color_list


def validate_colors(colors, colortype="tuple"):
    """
    Validates color(s) and returns a list of color(s) of a specified type
    """
    from numbers import Number

    if colors is None:
        colors = DEFAULT_PLOTLY_COLORS

    if isinstance(colors, str):
        if colors in PLOTLY_SCALES:
            colors_list = colorscale_to_colors(PLOTLY_SCALES[colors])
            # TODO: fix _gantt.py/_scatter.py so that they can accept the
            # actual colorscale and not just a list of the first and last
            # color in the plotly colorscale. In resolving this issue we
            # will be removing the immediate line below
            colors = [colors_list[0]] + [colors_list[-1]]
        elif "rgb" in colors or "#" in colors:
            colors = [colors]
        else:
            raise Exception(
                "If your colors variable is a string, it must be a "
                "Plotly scale, an rgb color or a hex color."
            )

    elif isinstance(colors, tuple):
        if isinstance(colors[0], Number):
            colors = [colors]
        else:
            colors = list(colors)

    # convert color elements in list to tuple color
    for j, each_color in enumerate(colors):
        if "rgb" in each_color:
            each_color = color_parser(each_color, unlabel_rgb)
            for value in each_color:
                if value > 255.0:
                    raise Exception(
                        "Whoops! The elements in your rgb colors "
                        "tuples cannot exceed 255.0."
                    )
            each_color = color_parser(each_color, unconvert_from_RGB_255)
            colors[j] = each_color

        if "#" in each_color:
            each_color = color_parser(each_color, hex_to_rgb)
            each_color = color_parser(each_color, unconvert_from_RGB_255)

            colors[j] = each_color

        if isinstance(each_color, tuple):
            for value in each_color:
                if value > 1.0:
                    raise Exception(
                        "Whoops! The elements in your colors tuples "
                        "cannot exceed 1.0."
                    )
            colors[j] = each_color

    if colortype == "rgb" and not isinstance(colors, str):
        for j, each_color in enumerate(colors):
            rgb_color = color_parser(each_color, convert_to_RGB_255)
            colors[j] = color_parser(rgb_color, label_rgb)

    return colors


def validate_colors_dict(colors, colortype="tuple"):
    """
    Validates dictionary of color(s)
    """
    # validate each color element in the dictionary
    for key in colors:
        if "rgb" in colors[key]:
            colors[key] = color_parser(colors[key], unlabel_rgb)
            for value in colors[key]:
                if value > 255.0:
                    raise Exception(
                        "Whoops! The elements in your rgb colors "
                        "tuples cannot exceed 255.0."
                    )
            colors[key] = color_parser(colors[key], unconvert_from_RGB_255)

        if "#" in colors[key]:
            colors[key] = color_parser(colors[key], hex_to_rgb)
            colors[key] = color_parser(colors[key], unconvert_from_RGB_255)

        if isinstance(colors[key], tuple):
            for value in colors[key]:
                if value > 1.0:
                    raise Exception(
                        "Whoops! The elements in your colors tuples "
                        "cannot exceed 1.0."
                    )

    if colortype == "rgb":
        for key in colors:
            colors[key] = color_parser(colors[key], convert_to_RGB_255)
            colors[key] = color_parser(colors[key], label_rgb)

    return colors


def convert_colors_to_same_type(
    colors,
    colortype="rgb",
    scale=None,
    return_default_colors=False,
    num_of_defualt_colors=2,
):
    """
    Converts color(s) to the specified color type

    Takes a single color or an iterable of colors, as well as a list of scale
    values, and outputs a 2-pair of the list of color(s) converted all to an
    rgb or tuple color type, aswell as the scale as the second element. If
    colors is a Plotly Scale name, then 'scale' will be forced to the scale
    from the respective colorscale and the colors in that colorscale will also
    be coverted to the selected colortype. If colors is None, then there is an
    option to return portion of the DEFAULT_PLOTLY_COLORS

    :param (str|tuple|list) colors: either a plotly scale name, an rgb or hex
        color, a color tuple or a list/tuple of colors
    :param (list) scale: see docs for validate_scale_values()

    :rtype (tuple) (colors_list, scale) if scale is None in the function call,
        then scale will remain None in the returned tuple
    """
    colors_list = []

    if colors is None and return_default_colors is True:
        colors_list = DEFAULT_PLOTLY_COLORS[0:num_of_defualt_colors]

    if isinstance(colors, str):
        if colors in PLOTLY_SCALES:
            colors_list = colorscale_to_colors(PLOTLY_SCALES[colors])
            if scale is None:
                scale = colorscale_to_scale(PLOTLY_SCALES[colors])

        elif "rgb" in colors or "#" in colors:
            colors_list = [colors]

    elif isinstance(colors, tuple):
        if isinstance(colors[0], Number):
            colors_list = [colors]
        else:
            colors_list = list(colors)

    elif isinstance(colors, list):
        colors_list = colors

    # validate scale
    if scale is not None:
        validate_scale_values(scale)

        if len(colors_list) != len(scale):
            raise Exception(
                "Make sure that the length of your scale matches the length "
                "of your list of colors which is {}.".format(len(colors_list))
            )

    # convert all colors to rgb
    for j, each_color in enumerate(colors_list):
        if "#" in each_color:
            each_color = color_parser(each_color, hex_to_rgb)
            each_color = color_parser(each_color, label_rgb)
            colors_list[j] = each_color

        elif isinstance(each_color, tuple):
            each_color = color_parser(each_color, convert_to_RGB_255)
            each_color = color_parser(each_color, label_rgb)
            colors_list[j] = each_color

    if colortype == "rgb":
        return (colors_list, scale)
    elif colortype == "tuple":
        for j, each_color in enumerate(colors_list):
            each_color = color_parser(each_color, unlabel_rgb)
            each_color = color_parser(each_color, unconvert_from_RGB_255)
            colors_list[j] = each_color
        return (colors_list, scale)
    else:
        raise Exception(
            "You must select either rgb or tuple for your colortype variable."
        )


def convert_dict_colors_to_same_type(colors_dict, colortype="rgb"):
    """
    Converts a colors in a dictionary of colors to the specified color type

    :param (dict) colors_dict: a dictionary whose values are single colors
    """
    for key in colors_dict:
        if "#" in colors_dict[key]:
            colors_dict[key] = color_parser(colors_dict[key], hex_to_rgb)
            colors_dict[key] = color_parser(colors_dict[key], label_rgb)

        elif isinstance(colors_dict[key], tuple):
            colors_dict[key] = color_parser(colors_dict[key], convert_to_RGB_255)
            colors_dict[key] = color_parser(colors_dict[key], label_rgb)

    if colortype == "rgb":
        return colors_dict
    elif colortype == "tuple":
        for key in colors_dict:
            colors_dict[key] = color_parser(colors_dict[key], unlabel_rgb)
            colors_dict[key] = color_parser(colors_dict[key], unconvert_from_RGB_255)
        return colors_dict
    else:
        raise Exception(
            "You must select either rgb or tuple for your colortype variable."
        )


def validate_scale_values(scale):
    """
    Validates scale values from a colorscale

    :param (list) scale: a strictly increasing list of floats that begins
        with 0 and ends with 1. Its usage derives from a colorscale which is
        a list of two-lists (a list with two elements) of the form
        [value, color] which are used to determine how interpolation weighting
        works between the colors in the colorscale. Therefore scale is just
        the extraction of these values from the two-lists in order
    """
    if len(scale) < 2:
        raise Exception(
            "You must input a list of scale values that has at least two values."
        )

    if (scale[0] != 0) or (scale[-1] != 1):
        raise Exception(
            "The first and last number in your scale must be 0.0 and 1.0 "
            "respectively."
        )

    if not all(x < y for x, y in zip(scale, scale[1:])):
        raise Exception(
            "'scale' must be a list that contains a strictly increasing "
            "sequence of numbers."
        )


def validate_colorscale(colorscale):
    """Validate the structure, scale values and colors of colorscale."""
    if not isinstance(colorscale, list):
        # TODO Write tests for these exceptions
        raise Exception("A valid colorscale must be a list.")
    if not all(isinstance(innerlist, list) for innerlist in colorscale):
        raise Exception("A valid colorscale must be a list of lists.")
    colorscale_colors = colorscale_to_colors(colorscale)
    scale_values = colorscale_to_scale(colorscale)

    validate_scale_values(scale_values)
    validate_colors(colorscale_colors)


def make_colorscale(colors, scale=None):
    """
    Makes a colorscale from a list of colors and a scale

    Takes a list of colors and scales and constructs a colorscale based
    on the colors in sequential order. If 'scale' is left empty, a linear-
    interpolated colorscale will be generated. If 'scale' is a specificed
    list, it must be the same legnth as colors and must contain all floats
    For documentation regarding to the form of the output, see
    https://plot.ly/python/reference/#mesh3d-colorscale

    :param (list) colors: a list of single colors
    """
    colorscale = []

    # validate minimum colors length of 2
    if len(colors) < 2:
        raise Exception("You must input a list of colors that has at least two colors.")

    if scale is None:
        scale_incr = 1.0 / (len(colors) - 1)
        return [[i * scale_incr, color] for i, color in enumerate(colors)]

    else:
        if len(colors) != len(scale):
            raise Exception("The length of colors and scale must be the same.")

        validate_scale_values(scale)

        colorscale = [list(tup) for tup in zip(scale, colors)]
        return colorscale


def find_intermediate_color(lowcolor, highcolor, intermed, colortype="tuple"):
    """
    Returns the color at a given distance between two colors

    This function takes two color tuples, where each element is between 0
    and 1, along with a value 0 < intermed < 1 and returns a color that is
    intermed-percent from lowcolor to highcolor. If colortype is set to 'rgb',
    the function will automatically convert the rgb type to a tuple, find the
    intermediate color and return it as an rgb color.
    """
    if colortype == "rgb":
        # convert to tuple color, eg. (1, 0.45, 0.7)
        lowcolor = unlabel_rgb(lowcolor)
        highcolor = unlabel_rgb(highcolor)

    diff_0 = float(highcolor[0] - lowcolor[0])
    diff_1 = float(highcolor[1] - lowcolor[1])
    diff_2 = float(highcolor[2] - lowcolor[2])

    inter_med_tuple = (
        lowcolor[0] + intermed * diff_0,
        lowcolor[1] + intermed * diff_1,
        lowcolor[2] + intermed * diff_2,
    )

    if colortype == "rgb":
        # back to an rgb string, e.g. rgb(30, 20, 10)
        inter_med_rgb = label_rgb(inter_med_tuple)
        return inter_med_rgb

    return inter_med_tuple


def unconvert_from_RGB_255(colors):
    """
    Return a tuple where each element gets divided by 255

    Takes a (list of) color tuple(s) where each element is between 0 and
    255. Returns the same tuples where each tuple element is normalized to
    a value between 0 and 1
    """
    return (colors[0] / (255.0), colors[1] / (255.0), colors[2] / (255.0))


def convert_to_RGB_255(colors):
    """
    Multiplies each element of a triplet by 255

    Each coordinate of the color tuple is rounded to the nearest float and
    then is turned into an integer. If a number is of the form x.5, then
    if x is odd, the number rounds up to (x+1). Otherwise, it rounds down
    to just x. This is the way rounding works in Python 3 and in current
    statistical analysis to avoid rounding bias

    :param (list) rgb_components: grabs the three R, G and B values to be
        returned as computed in the function
    """
    rgb_components = []

    for component in colors:
        rounded_num = decimal.Decimal(str(component * 255.0)).quantize(
            decimal.Decimal("1"), rounding=decimal.ROUND_HALF_EVEN
        )
        # convert rounded number to an integer from 'Decimal' form
        rounded_num = int(rounded_num)
        rgb_components.append(rounded_num)

    return (rgb_components[0], rgb_components[1], rgb_components[2])


def n_colors(lowcolor, highcolor, n_colors, colortype="tuple"):
    """
    Splits a low and high color into a list of n_colors colors in it

    Accepts two color tuples and returns a list of n_colors colors
    which form the intermediate colors between lowcolor and highcolor
    from linearly interpolating through RGB space. If colortype is 'rgb'
    the function will return a list of colors in the same form.
    """
    if colortype == "rgb":
        # convert to tuple
        lowcolor = unlabel_rgb(lowcolor)
        highcolor = unlabel_rgb(highcolor)

    diff_0 = float(highcolor[0] - lowcolor[0])
    incr_0 = diff_0 / (n_colors - 1)
    diff_1 = float(highcolor[1] - lowcolor[1])
    incr_1 = diff_1 / (n_colors - 1)
    diff_2 = float(highcolor[2] - lowcolor[2])
    incr_2 = diff_2 / (n_colors - 1)
    list_of_colors = []

    for index in range(n_colors):
        new_tuple = (
            lowcolor[0] + (index * incr_0),
            lowcolor[1] + (index * incr_1),
            lowcolor[2] + (index * incr_2),
        )
        list_of_colors.append(new_tuple)

    if colortype == "rgb":
        # back to an rgb string
        list_of_colors = color_parser(list_of_colors, label_rgb)

    return list_of_colors


def label_rgb(colors):
    """
    Takes tuple (a, b, c) and returns an rgb color 'rgb(a, b, c)'
    """
    return "rgb(%s, %s, %s)" % (colors[0], colors[1], colors[2])


def unlabel_rgb(colors):
    """
    Takes rgb color(s) 'rgb(a, b, c)' and returns tuple(s) (a, b, c)

    This function takes either an 'rgb(a, b, c)' color or a list of
    such colors and returns the color tuples in tuple(s) (a, b, c)
    """
    str_vals = ""
    for index in range(len(colors)):
        try:
            float(colors[index])
            str_vals = str_vals + colors[index]
        except ValueError:
            if colors[index] == "," or colors[index] == ".":
                str_vals = str_vals + colors[index]

    str_vals = str_vals + ","
    numbers = []
    str_num = ""
    for char in str_vals:
        if char != ",":
            str_num = str_num + char
        else:
            numbers.append(float(str_num))
            str_num = ""
    return (numbers[0], numbers[1], numbers[2])


def hex_to_rgb(value):
    """
    Calculates rgb values from a hex color code.

    :param (string) value: Hex color string

    :rtype (tuple) (r_value, g_value, b_value): tuple of rgb values
    """
    value = value.lstrip("#")
    hex_total_length = len(value)
    rgb_section_length = hex_total_length // 3
    return tuple(
        int(value[i : i + rgb_section_length], 16)
        for i in range(0, hex_total_length, rgb_section_length)
    )


def colorscale_to_colors(colorscale):
    """
    Extracts the colors from colorscale as a list
    """
    color_list = []
    for item in colorscale:
        color_list.append(item[1])
    return color_list


def colorscale_to_scale(colorscale):
    """
    Extracts the interpolation scale values from colorscale as a list
    """
    scale_list = []
    for item in colorscale:
        scale_list.append(item[0])
    return scale_list


def convert_colorscale_to_rgb(colorscale):
    """
    Converts the colors in a colorscale to rgb colors

    A colorscale is an array of arrays, each with a numeric value as the
    first item and a color as the second. This function specifically is
    converting a colorscale with tuple colors (each coordinate between 0
    and 1) into a colorscale with the colors transformed into rgb colors
    """
    for color in colorscale:
        color[1] = convert_to_RGB_255(color[1])

    for color in colorscale:
        color[1] = label_rgb(color[1])
    return colorscale


def named_colorscales():
    """
    Returns lowercased names of built-in continuous colorscales.
    """
    from _plotly_utils.basevalidators import ColorscaleValidator

    return [c for c in ColorscaleValidator("", "").named_colorscales]


def get_colorscale(name):
    """
    Returns the colorscale for a given name. See `named_colorscales` for the
    built-in colorscales.
    """
    from _plotly_utils.basevalidators import ColorscaleValidator

    if not isinstance(name, str):
        raise Exception("Name argument have to be a string.")

    name = name.lower()
    if name[-2:] == "_r":
        should_reverse = True
        name = name[:-2]
    else:
        should_reverse = False

    if name in ColorscaleValidator("", "").named_colorscales:
        colorscale = ColorscaleValidator("", "").named_colorscales[name]
    else:
        raise Exception(f"Colorscale {name} is not a built-in scale.")

    if should_reverse:
        colorscale = colorscale[::-1]
    return make_colorscale(colorscale)


def sample_colorscale(colorscale, samplepoints, low=0.0, high=1.0, colortype="rgb"):
    """
    Samples a colorscale at specific points.

    Interpolates between colors in a colorscale to find the specific colors
    corresponding to the specified sample values. The colorscale can be specified
    as a list of `[scale, color]` pairs, as a list of colors, or as a named
    plotly colorscale. The samplepoints can be specefied as an iterable of specific
    points in the range [0.0, 1.0], or as an integer number of points which will
    be spaced equally between the low value (default 0.0) and the high value
    (default 1.0). The output is a list of colors, formatted according to the
    specified colortype.
    """
    from bisect import bisect_left

    try:
        validate_colorscale(colorscale)
    except Exception:
        if isinstance(colorscale, str):
            colorscale = get_colorscale(colorscale)
        else:
            colorscale = make_colorscale(colorscale)

    scale = colorscale_to_scale(colorscale)
    validate_scale_values(scale)
    colors = colorscale_to_colors(colorscale)
    colors = validate_colors(colors, colortype="tuple")

    if isinstance(samplepoints, int):
        samplepoints = [
            low + idx / (samplepoints - 1) * (high - low) for idx in range(samplepoints)
        ]
    elif isinstance(samplepoints, float):
        samplepoints = [samplepoints]

    sampled_colors = []
    for point in samplepoints:
        high = bisect_left(scale, point)
        low = high - 1
        interpolant = (point - scale[low]) / (scale[high] - scale[low])
        sampled_color = find_intermediate_color(colors[low], colors[high], interpolant)
        sampled_colors.append(sampled_color)
    return validate_colors(sampled_colors, colortype=colortype)
